# -*- coding: utf-8 -*-


if __name__ == "__main__":
    import os
    import pathlib
    os.chdir(pathlib.Path(__file__).parent.parent.parent.resolve())     
    print('__main__ - current working directory:', os.getcwd())
    
from frm.frm.market_data.fx_volatility_surface_helpers import fx_σ_input_helper     
from frm.frm.market_data.ir_zero_curve import ZeroCurve
from frm.frm.pricing_engine.garman_kohlhagen import gk_price, gk_solve_implied_σ, gk_solve_strike
from frm.frm.pricing_engine.heston_gk import heston_fit_vanilla_fx_smile, heston1993_price_fx_vanilla_european, heston_carr_madan_fx_vanilla_european

from frm.frm.schedule.tenor import calc_tenor_date, get_spot_offset
from frm.frm.schedule.daycounter import DayCounter, VALID_DAY_COUNT_BASIS_TYPES
from frm.frm.schedule.business_day_calendar import get_calendar        
from frm.frm.utilities.utilities import convert_column_type    

import numpy as np
import pandas as pd
import datetime as dt
from scipy.optimize import fsolve, root_scalar
from scipy.interpolate import CubicSpline, InterpolatedUnivariateSpline 
from dataclasses import dataclass, field, InitVar
from typing import Optional, Union, Literal
import warnings
import re
import matplotlib.pyplot as plt

VALID_DELTA_CONVENTIONS = ['regular_spot_delta','regular_forward_delta','premium_adjusted_spot_delta','premium_adjusted_forward_delta'] 
VALID_FX_SMILE_INTERPOLATION_METHOD = Literal['univariate_spline',
                                              'cubic_spline',
                                              'heston_analytical_1993',
                                              'heston_carr_madan_gauss_kronrod_quadrature',
                                              'heston_carr_madan_fft_w_simpsons',
                                              'vanna_volga',
                                              'sabr',
                                              'svi_sabr']





@dataclass
class FXVolatilitySurface:
    curve_date: pd.Timestamp
    curve_ccy: str
    fx_forward_curve: pd.DataFrame
    domestic_zero_curve: ZeroCurve
    foreign_zero_curve: ZeroCurve

    # Non initialisation arguments
    daycounter: DayCounter = field(init=False)
    spot_date: pd.Timestamp
    fx_spot_rate: float = field(init=False)
    domestic_ccy: str = field(init=False)
    foreign_ccy: str = field(init=False)
    Δ_σ_daily: pd.DataFrame=None
    σ_pillar: dict=None
    K_σ_daily_smile_func: dict=None
    
    # volatility definitions
    σ_pillar: pd.DataFrame=None
    smile_interpolation_method: Optional[VALID_FX_SMILE_INTERPOLATION_METHOD] = 'cubic_spline'
    spot_date: Optional[pd.Timestamp] = None
    
    #atmΔ: pd.DataFrame=None
    #atmF: pd.DataFrame=None
    #bf25Δ: pd.DataFrame=None
    #rr25Δ: pd.DataFrame=None
    #bf10Δ: pd.DataFrame=None
    #bf10Δ: pd.DataFrame=None
    
    day_count_basis: InitVar[Optional[VALID_DAY_COUNT_BASIS_TYPES]] = 'act/act'
    
    

    def __post_init__(self, day_count_basis):
         
        self.daycounter = DayCounter(day_count_basis)
        self.K_σ_daily_smile_func = dict()

        holiday_calendar = get_calendar(ccys=[self.curve_ccy[:3],self.curve_ccy[3:]])

        if self.spot_date is None:
            result = calc_tenor_date(self.curve_date, 'sp', self.curve_ccy, holiday_calendar=holiday_calendar, spot_offset=True)
            holiday_rolled_offset_date, cleaned_tenor_name, spot_date = result
            assert holiday_rolled_offset_date == spot_date
            self.spot_date = spot_date

        # curve_ccy validation
        self.curve_ccy = self.curve_ccy.lower().strip()
        assert len(self.curve_ccy) == 6, self.curve_ccy
        self.foreign_ccy = self.curve_ccy[:3]
        self.domestic_ccy = self.curve_ccy[3:]
        if self.foreign_ccy == 'usd' and self.domestic_ccy in {'aud','eur','gbp','nzd'}:
            warnings.warn("non conventional fx market, typically 'usd' is the domestic currency for 'audusd', 'eurusd', 'gbpusd' and 'nzdusd' pairs")
        elif self.domestic_ccy == 'usd' and self.foreign_ccy not in {'aud','eur','gbp','nzd'}:
            warnings.warn("non conventional fx market, typically 'usd' is the foreign currency  except for 'audusd', 'eurusd', 'gbpusd' and 'nzdusd' pairs")
 
        # fx_forward_curve validation 
        assert 'fx_forward_rate' in self.fx_forward_curve.columns 
        self.fx_forward_curve = convert_column_type(self.fx_forward_curve)
        assert ('tenor_name' in self.fx_forward_curve.columns) or ('expiry_date' in self.fx_forward_curve.columns and 'delivery_date' in self.fx_forward_curve.columns)
        if 'expiry_date' not in self.fx_forward_curve.columns and 'delivery_date' not in self.fx_forward_curve.columns:
            
            result = calc_tenor_date(self.curve_date, self.fx_forward_curve['tenor_name'], self.curve_ccy, holiday_calendar=holiday_calendar, spot_offset=True)
            holiday_rolled_offset_date, cleaned_tenor_name, spot_date = result
            self.fx_forward_curve['tenor_name'] = cleaned_tenor_name
            self.fx_forward_curve['delivery_date'] = holiday_rolled_offset_date
            self.fx_forward_curve['years_to_delivery'] = self.daycounter.year_fraction(self.curve_date, self.fx_forward_curve['tenor_years_delivery'])            

            spot_offset = -1 * get_spot_offset(curve_ccy=self.curve_ccy)
            holiday_rolled_offset_date = np.busday_offset(holiday_rolled_offset_date, offsets=spot_offset, roll='preceeding', busdaycal=holiday_calendar)
                
            result = calc_tenor_date(self.curve_date, self.fx_forward_curve['tenor_name'], self.curve_ccy, holiday_calendar=holiday_calendar, spot_offset=True)
            holiday_rolled_offset_date, cleaned_tenor_name, spot_date = result
            
            self.fx_forward_curve['expiry_date'] = holiday_rolled_offset_date
            self.fx_forward_curve['yrs_to_expiry'] = self.daycounter.year_fraction(self.curve_date, self.fx_forward_curve['tenor_years_observation'])
                                    
        self.fx_forward_curve.set_index('expiry_date', inplace=True,drop=True)
            
        if self.curve_date not in self.fx_forward_curve.index:
            raise ValueError("The curve date (" + self.curve_date.strftime('%Y-%m-%d') + ') fx rate (i.e the fx spot rate) is in missing in fx_forward_curve')
        
        self.fx_spot_rate = self.fx_forward_curve.loc[self.curve_date,'fx_forward_rate']
        
        # σ_pillar validation
        if self.σ_pillar is not None:
            
            σ_pillar = self.σ_pillar
            
            σ_pillar['curve_date'] = self.curve_date
            σ_pillar['curve_ccy'] = self.curve_ccy
            
            
            results = fx_σ_input_helper(σ_pillar)
            
            
            # Convert columns to type float (can be set to type object from splicing or NaNs)
            for col in self.σ_pillar.columns:
                if self.σ_pillar[col].apply(isinstance, args=(float,)).all():
                    self.σ_pillar[col] = pd.to_numeric(self.σ_pillar[col])
            if 'tenor_date' not in self.σ_pillar.columns:
                result = calc_tenor_date(self.curve_date, self.σ_pillar['tenor_name'], self.curve_ccy, holiday_calendar=holiday_calendar)
                holiday_rolled_offset_date, cleaned_tenor_name, spot_date = result            
                self.σ_pillar['tenor_date'] = holiday_rolled_offset_date
                self.σ_pillar['tenor_name'] = cleaned_tenor_name
            self.σ_pillar.set_index('tenor_date', inplace=True,drop=True)       
            self.σ_pillar.insert(loc=0, column='tenor_years', value=self.daycounter.year_fraction(self.curve_date, self.σ_pillar.index).values)                    

            cols = [v for v in self.σ_pillar.columns if v not in ['tenor_date','tenor_name','tenor_years','Δ_convention']]
            for i,v in enumerate(cols):
                pattern1 = r'^σ_(\d{1,2})Δ(call|put)$'
                pattern2 = r'^σ_(\d{1,2})Δ(bf|rr)$'
                atm_column_names = ['σ_atmΔneutral','σ_atmf']
                if (re.match(pattern1, v) and 1 <= int(re.match(pattern1, v).group(1)) <= 99) \
                    or (re.match(pattern2, v) and 1 <= int(re.match(pattern2, v).group(1)) <= 99) \
                    or v in atm_column_names:    
                    pass
                else:
                    msg = 'user added column' + "'" + v + "'" + ' does not ' \
                        + 'match regex pattern ' + "'" + pattern1 + "', or pattern " + "'" + pattern2 + "'," \
                        + ' and is not in the allowed list (' + ', '.join(atm_column_names) + ')'
                    raise ValueError(msg)
            self.σ_pillar = self.__interp_daily_FXF_and_IR_rates(self.σ_pillar.copy())
            
            
            if True:                
                ###################################################################
                # 1. Calculate the delta-strike surface from the delta-volatility surface           
                
                self.σ_pillar_input_Δ_convention = self.σ_pillar.copy()
                Δ_K_pillar = self.σ_pillar.copy()
                Δ_K_pillar.loc[:, Δ_K_pillar.columns.str.contains('σ')] = np.nan
                Δ_K_pillar.rename(columns=lambda x: x.replace('σ', 'k'), inplace=True)
    
                helper_cols = ['tenor_date','tenor_name','tenor_years','fx_forward_rate','foreign_ccy_continuously_compounded_zero_rate','domestic_ccy_continuously_compounded_zero_rate','Δ_convention']
                dict_K_σ = {v.replace('σ', 'k'): v for v in self.σ_pillar.columns if v not in helper_cols}
                for date, row in Δ_K_pillar.iterrows():
                    
                    
                    
                    # Solve strikes 
                    
                    # Scalars
                    S = self.fx_spot_rate
                    r_f=row['foreign_ccy_continuously_compounded_zero_rate']
                    r_d=row['domestic_ccy_continuously_compounded_zero_rate']
                    tau=row['tenor_years']                
                    Δ_convention=row['Δ_convention']
                    F=row['fx_forward_rate']  
                    # Arrays
                    cp = [1 if v[-4:] == 'call' else -1 if v[-3:] == 'put' else None for v in dict_K_σ.keys()]
                    Δ = [0.5 if v == 'k_atmΔneutral' else cp[i] * float(v.split('_')[1].split('Δ')[0]) / 100 for i,v in enumerate(dict_K_σ.keys())]
                    σ = self.σ_pillar.loc[date, dict_K_σ.values()].values
                                        
                    K = gk_solve_strike(S=S,tau=tau,r_f=r_f,r_d=r_d,σ=σ,Δ=Δ,Δ_convention=Δ_convention,F=F)
                    Δ_K_pillar.loc[date,dict_K_σ.keys()] = K
                                                
  
                self.Δ_K_pillar = Δ_K_pillar

            ###################################################################
            # Interpolate the term structure of volatility to get a daily volatility smile            
            dates = pd.date_range(min(self.σ_pillar.index),max(self.σ_pillar.index),freq='d')
            tenor_years = self.daycounter.year_fraction(self.curve_date,dates)
            df_interp = pd.DataFrame({'tenor_date':dates, 'tenor_years': tenor_years})
            df_pillar = self.σ_pillar.copy()

            # Merge to find closest smaller and larger tenors for each target tenor
            df_lower_pillar = pd.merge_asof(df_interp.sort_values('tenor_years'), df_pillar, on='tenor_date', direction='backward', suffixes=('_interp', '_pillar'))
            df_upper_pillar = pd.merge_asof(df_interp.sort_values('tenor_years'), df_pillar, on='tenor_date', direction='forward', suffixes=('_interp', '_pillar'))
                    
            # Convert to numpy for efficient calculations
            t1 = df_lower_pillar['tenor_years_pillar'].to_numpy()
            t2 = df_upper_pillar['tenor_years_pillar'].to_numpy()
            t = df_interp['tenor_years'].to_numpy()
            t1 = t1[:, np.newaxis]
            t2 = t2[:, np.newaxis]
            t = t[:, np.newaxis]
            σ_t1 = df_lower_pillar[cols].to_numpy()
            σ_t2 = df_upper_pillar[cols].to_numpy()             
            
            # Interpolation logic
            mask = t1 != t2
            σ_t = np.where(mask, self.flat_forward_interp(t1, σ_t1, t2, σ_t2, t), σ_t1)
            df_interp[cols] = σ_t            
            df_interp['Δ_convention'] = 'regular_forward_Δ' # need to add a section to calculate forward delta from spot delta
            df_interp.set_index('tenor_date', inplace=True,drop=True)    
            self.Δ_σ_daily = self.__interp_daily_FXF_and_IR_rates(df_interp)


    def __interp_daily_FXF_and_IR_rates(self, df):
        offset = 1 if 'tenor_name' in df.columns else 0
        
        df.insert(loc=1+offset, column='foreign_ccy_continuously_compounded_zero_rate', value=np.nan)
        df.insert(loc=2+offset, column='domestic_ccy_continuously_compounded_zero_rate', value=np.nan)
        df.insert(loc=3+offset, column='fx_forward_rate', value=np.nan)
    
        df['foreign_ccy_continuously_compounded_zero_rate'] = self.foreign_zero_curve.zero_rate(dates=df.index,compounding_frequency='continuously').values       
        df['domestic_ccy_continuously_compounded_zero_rate'] = self.domestic_zero_curve.zero_rate(dates=df.index,compounding_frequency='continuously').values
        df['fx_forward_rate'] = self.interp_fx_forward_curve(expiry_dates=df.index, flat_extrapolation=True)
        return df


    def interp_fx_forward_curve(self, 
                                expiry_dates: pd.DatetimeIndex,
                                flat_extrapolation: bool=True):
        """
        Interpolate the FX forward curve to match given expiry_dates.
        Please note expiry date is 2 business days prior to the spot date
        """
        
        unique_dates = expiry_dates.drop_duplicates()
        combined_index = self.fx_forward_curve.index.union(unique_dates)
        result = self.fx_forward_curve.reindex(combined_index)
        start_date, end_date = self.fx_forward_curve.index.min(), self.fx_forward_curve.index.max()
        
        if flat_extrapolation:
            result['fx_forward_rate'] = result['fx_forward_rate'].interpolate(method='time', limit_area='inside').ffill().bfill()
            # Find out of range dates and warn
            out_of_range_dates = unique_dates[(unique_dates < start_date) | (unique_dates > end_date)]
            for date in out_of_range_dates:
                warnings.warn(f"Date {date} is outside the range {start_date} - {end_date}, flat extrapolation applied.")
        else:
            result['fx_forward_rate'] = result['fx_forward_rate'].interpolate(method='time', limit_area='inside')
        
        result = result.reindex(expiry_dates)
        
        return result['fx_forward_rate']
        
        
                
    def forward_volatility(self, 
                           t1: Union[float, np.array], 
                           σ_t1: Union[float, np.array], 
                           t2: Union[float, np.array], 
                           σ_t2: Union[float, np.array]) -> Union[float, np.array]:
        """
        Calculate the at-the-money forward volatility from time t1 to t2.
        The forward volatility is based on the consistency condition:
        σ_t1**2 * t1 + σ_t1_t2**2 * (t2- t1) = σ_t1**2 * t2    
    
        Parameters:
        - t1 (float): Time to first maturity
        - σ_t1 (float): At-the-money volatility at time t1
        - t2 (float): Time to second maturity
        - σ_t2 (float): At-the-money volatility at time t2
    
        Returns:
        - np.array: Forward volatility from time t1 to t2
        """
        tau = t2 - t1
        if np.any(tau == 0):
            warnings.warn("t2 and t1 are equal. NaN will be returned.")
    
        result = (σ_t2**2 * t2 - σ_t1**2 * t1) / tau
        if np.any(result < 0):
            raise ValueError("Negative value encountered under square root.")
    
        return np.sqrt(result)

    def flat_forward_interp(self, 
                            t1: Union[float, np.array], 
                            σ_t1: Union[float, np.array], 
                            t2: Union[float, np.array], 
                            σ_t2: Union[float, np.array], 
                            t: Union[float, np.array]) -> Union[float, np.array]:
        """
        Interpolate the at-the-money volatility at a given time 't' using flat forward interpolation.
    
        Parameters:
        - t1 (Union[float, array]): Time to first maturity
        - σ_t1 (Union[float, array]): At-the-money volatility at time t1
        - t2 (Union[float, array]): Time to second maturity
        - σ_t2 (Union[float, array]): At-the-money volatility at time t2
        - t (Union[float, array]): Time at which to interpolate the volatility
    
        Returns:
        - array: Interpolated volatility at time 't'
        """
        σ_t12 = self.forward_volatility(t1, σ_t1, t2, σ_t2)
        return np.sqrt((σ_t1**2 * t1 + σ_t12**2 * (t - t1)) / t)



    def fit_and_plot_smile(self):
        
        def convert_colunm_name_to_delta_for_plt(input_str):
            if 'put' in input_str:
                return int(input_str.split('Δ')[0][2:])
            elif 'call' in input_str:
                return 100 - int(input_str.split('Δ')[0][2:])
            elif 'neutral' in input_str:
                return 50
        
        def convert_colunm_name_to_delta(input_str):
            if 'put' in input_str:
                return -1 * int(input_str.split('Δ')[0][2:]) / 100
            elif 'call' in input_str:
                return int(input_str.split('Δ')[0][2:]) / 100
            elif 'neutral' in input_str:
                return 0.5           
            
        i = 0
        for date, row in self.σ_pillar.iterrows():
            
            # Scalars
            tenor_name = row['tenor_name']
            Δ_convention=row['Δ_convention']
            S = self.fx_spot_rate
            r_f=row['foreign_ccy_continuously_compounded_zero_rate']
            r_d=row['domestic_ccy_continuously_compounded_zero_rate']
            tau=row['tenor_years']                
            F=row['fx_forward_rate'] 
            
            # Arrays
            Δ = [convert_colunm_name_to_delta(c) for c in self.σ_pillar.columns if c[0] == 'σ']
            Δ_for_plt = [convert_colunm_name_to_delta_for_plt(c) for c in self.σ_pillar.columns if c[0] == 'σ']
            cp = np.sign(Δ)
            σ = row[[c for c in self.σ_pillar.columns if c[0] == 'σ']].values             
                     
            if self.smile_interpolation_method[:6] == 'heston':
                v0, vv, kappa, theta, rho, lambda_, IV, SSE = \
                    heston_fit_vanilla_fx_smile(Δ, Δ_convention, σ, S, r_f, r_d, tau, cp, pricing_method=self.smile_interpolation_method)
            
                # Displaying output
                print(f'=== {tenor_name} calibration results ===')
                print(f'v0, vv, kappa, theta, rho: {v0, vv, kappa, theta, rho}')
                print(f'SSE {SSE*100}')
                
                # Plotting
                i+=1
                plt.figure(i+1)
                plt.plot(Δ_for_plt, σ * 100, 'ko-', linewidth=1)
                plt.plot(Δ_for_plt, IV * 100, 'rs--', linewidth=1)
                plt.legend([f'{tenor_name} smile', 'Heston fit'], loc='upper right')
                plt.xlabel('Delta [%]')
                plt.ylabel('Implied volatility [%]')
                plt.xticks(Δ_for_plt)
                plt.title(self.smile_interpolation_method)
                plt.show()   
                
            else:
                raise ValueError
                  

    def interp_σ_surface(self, dates: pd.DatetimeIndex, K: np.array, cp):
        """
        Private method to interpolate the σ surface for given dates, updating the 
        `K_σ_daily_smile_func` attribute with appropriate interpolation functions or Heston parameters.
    
        Parameters:
        dates (pd.DatetimeIndex): Array of dates to perform the σ surface interpolation for.
    
        Raises:
        ValueError: If the sum of squared errors (SSE) from the Heston fit exceeds a threshold, indicating a poor fit.
    
        Note:
        Updates the `K_σ_daily_smile_func` attribute with either a spline interpolation function or Heston parameters for each date in `dates`.
        """        
        
        K = np.atleast_1d(K).astype(float)
        cp = np.atleast_1d(cp).astype(float)
        
        assert dates.shape == K.shape
        assert dates.shape == cp.shape
        
        df = self.Δ_σ_daily.loc[pd.DatetimeIndex(dates),:].copy()
        cols_to_drop = [col for col in df.columns if 'σ' in col]
        df.drop(columns=cols_to_drop, inplace=True)
        df['K'] = K
        df['call_put'] = cp

        df_unique_dates = self.Δ_σ_daily.loc[pd.DatetimeIndex(set(dates)),:].copy()
        df_unique_dates.loc[:, df_unique_dates.columns.str.contains('σ')] = np.nan
        df_unique_dates.rename(columns=lambda x: x.replace('σ', 'k'), inplace=True)

        helper_cols = ['tenor_date','tenor_name','tenor_years','fx_forward_rate','foreign_ccy_continuously_compounded_zero_rate','domestic_ccy_continuously_compounded_zero_rate','Δ_convention']
        dict_K_σ = {v.replace('σ', 'k'): v for v in self.Δ_σ_daily.columns if v not in helper_cols}
                
        result = []
    
        # Interpolate volatility surface for each date in df
        for date, row in df_unique_dates.iterrows():
            
            if date not in self.K_σ_daily_smile_func.keys(): 
                
                # Scalars
                S = self.fx_spot_rate
                r_f=row['foreign_ccy_continuously_compounded_zero_rate']
                r_d=row['domestic_ccy_continuously_compounded_zero_rate']
                tau=row['tenor_years']                
                Δ_convention=row['Δ_convention']
                F=row['fx_forward_rate']     
                
                # Arrays
                cp = [1 if v[-4:] == 'call' else -1 if v[-3:] == 'put' else 1 for v in dict_K_σ.keys()]
                Δ = [0.5 if v == 'k_atmΔneutral' else cp[i] * float(v.split('_')[1].split('Δ')[0]) / 100 for i,v in enumerate(dict_K_σ.keys())]
                σ = self.Δ_σ_daily.loc[date, dict_K_σ.values()].values                
                            
                if self.smile_interpolation_method == 'univariate_spline':
                    K = gk_solve_strike(S=S,tau=tau,r_f=r_f,r_d=r_d,σ=σ,Δ=Δ,Δ_convention=Δ_convention,F=F)
                    df_unique_dates.loc[date,dict_K_σ.keys()] = K 
                    self.K_σ_daily_smile_func[date] = InterpolatedUnivariateSpline(x=K, y=σ)
                
                elif self.smile_interpolation_method == 'cubic_spline':
                    K = gk_solve_strike(S=S,tau=tau,r_f=r_f,r_d=r_d,σ=σ,Δ=Δ,Δ_convention=Δ_convention,F=F)
                    df_unique_dates.loc[date,dict_K_σ.keys()] = K 
                    self.K_σ_daily_smile_func[date] = CubicSpline(x=K, y=σ)
                    
                elif self.smile_interpolation_method[:6] == 'heston':
                    v0, vv, kappa, theta, rho, lambda_, IV, SSE = \
                        heston_fit_vanilla_fx_smile(Δ, Δ_convention, σ, S, r_f, r_d, tau, cp, pricing_method=self.smile_interpolation_method)
                    if SSE < 0.001:
                        self.K_σ_daily_smile_func[date] = v0, vv, kappa, theta, rho, lambda_, IV, SSE
                    else:
                        raise ValueError('SSE is a large value, ', round(SSE,4), ' heston fit at ', date,' is likely poor')
            
        # Get the implied vol for each strike and date combination
        for date,row in df.iterrows():
            K_target = row['K']
            
            if self.smile_interpolation_method in {'univariate_spline', 'cubic_spline'}:
                σ = self.K_σ_daily_smile_func[date](K_target)
                result.append(σ)
            elif self.smile_interpolation_method[:6] == 'heston':
                S = self.fx_spot_rate
                r_f=row['foreign_ccy_continuously_compounded_zero_rate']
                r_d=row['domestic_ccy_continuously_compounded_zero_rate']
                tau=row['tenor_years']                
                F=row['fx_forward_rate']               
                cp = row['call_put']

                v0, vv, kappa, theta, rho, lambda_, IV, SSE = self.K_σ_daily_smile_func[date]
                
                if self.smile_interpolation_method == 'heston_analytical_1993':
                    X = heston1993_price_fx_vanilla_european(S, tau, r_f, r_d, cp, K_target, v0, vv, kappa, theta, rho, lambda_)
                elif self.smile_interpolation_method == 'heston_carr_madan_gauss_kronrod_quadrature':     
                    X = heston_carr_madan_fx_vanilla_european(S, tau, r_f, r_d, cp, K_target, v0, vv, kappa, theta, rho, integration_method=0)
                elif self.smile_interpolation_method == 'heston_carr_madan_fft_w_simpsons':
                    X = heston_carr_madan_fx_vanilla_european(S, tau, r_f, r_d, cp, K_target, v0, vv, kappa, theta, rho, integration_method=1)
                
                implied_σ = gk_solve_implied_σ(S=S, tau=tau, r_f=r_f, r_d=r_d, cp=cp, K=K_target, X=X, σ_guess=v0) 
                result.append(implied_σ)            
                        
        return np.array(result)
    
    def price_fx_vanilla_european(self, 
                                  expiry_datetimeindex,        
                                  K, 
                                  cp,
                                  analytical_greeks_flag):
                      
        K = np.atleast_1d(K)
        cp = np.atleast_1d(cp)

        assert expiry_datetimeindex.shape == K.shape
        assert expiry_datetimeindex.shape == cp.shape

        σ = self.interp_σ_surface(expiry_datetimeindex, K, cp)
        r_f = self.foreign_zero_curve.zero_rate(expiry_datetimeindex,compounding_frequency='continuously').values
        r_d = self.domestic_zero_curve.zero_rate(expiry_datetimeindex,compounding_frequency='continuously').values
        F = self.interp_fx_forward_curve(expiry_dates=expiry_datetimeindex, flat_extrapolation=True)

        result = gk_price(
            S=self.fx_spot_rate,
            tau=self.daycounter.year_fraction(self.curve_date, expiry_datetimeindex),
            r_f=r_f,
            r_d=r_d,
            cp=cp,
            K=K,
            σ=σ,
            F=F,
            analytical_greeks_flag=analytical_greeks_flag
        )    

        if analytical_greeks_flag:
            return result[0], result[1], pd.DataFrame({'σ': σ, 'r_f':r_f, 'r_d':r_d, 'F':F})

        return result
        







# -*- coding: utf-8 -*-


if __name__ == "__main__":
    import os
    import pathlib
    os.chdir(pathlib.Path(__file__).parent.parent.parent.resolve())     
    print('__main__ - current working directory:', os.getcwd())
    
from frm.frm.market_data.ir_zero_curve import ZeroCurve
from frm.frm.pricing_engine.garman_kohlhagen import gk_price, gk_solve_implied_σ, gk_solve_strike
from frm.frm.pricing_engine.heston_garman_kohlhagen import heston_fit_vanilla_fx_smile, heston1993_price_fx_vanilla_european, heston_carr_madan_price_fx_vanilla_european

from frm.frm.schedule.tenor import calc_tenor_date, get_spot_offset
from frm.frm.schedule.daycounter import VALID_DAY_COUNT_BASIS
from frm.frm.schedule.business_day_calendar import get_calendar        
from frm.frm.utilities.utilities import convert_column_to_consistent_data_type, clean_input_dataframe, move_col_after, copy_errors_and_warnings_to_input

import numpy as np
import pandas as pd
import datetime as dt
from scipy.optimize import fsolve, root_scalar
from scipy.interpolate import CubicSpline, InterpolatedUnivariateSpline 
from dataclasses import dataclass, field, InitVar
from typing import Optional, Union, Literal
import warnings
import re
import matplotlib.pyplot as plt

VALID_DELTA_CONVENTIONS = ['regular_spot_delta','regular_forward_delta','premium_adjusted_spot_delta','premium_adjusted_forward_delta'] 


def interp_fx_forward_curve(fx_forward_curve: pd.DataFrame, 
                            dates: Union[pd.Series, pd.DatetimeIndex]=None,
                            date_type: str=None,
                            flat_extrapolation: bool=True) -> pd.Series:
    
    assert date_type in ['expiry_date','delivery_date']
    
    if isinstance(dates, pd.Series):
        dates = pd.DatetimeIndex(dates)
    elif isinstance(dates, pd.DatetimeIndex):
        pass
    else:
        raise ValueError("'dates' must be a pandas Series or DatetimeIndex")
        
    unique_dates = dates.drop_duplicates()
    combined_dates = unique_dates.union(pd.DatetimeIndex(fx_forward_curve[date_type]))
    df = pd.Series(fx_forward_curve['fx_forward_rate'].values, index=fx_forward_curve[date_type].values)
    
    result = df.reindex(combined_dates.values).copy()
    start_date = fx_forward_curve[date_type].min()
    end_date = fx_forward_curve[date_type].max()
    
    if flat_extrapolation:
        try:                    
            result = result.interpolate(method='time', limit_area='inside').ffill().bfill()
        except:
            pass
        # Find out of range dates and warn
        out_of_range_dates = unique_dates[(unique_dates < start_date) | (unique_dates > end_date)]
        for date in out_of_range_dates:
            warnings.warn(f"Date {date} is outside the {date_type} range {start_date} - {end_date}, flat extrapolation applied.")
    else:
        result = result.interpolate(method='time', limit_area='inside')
    
    result = result.reindex(dates)
    
    return result
    

def fx_σ_input_helper(df):
    
    df_input = df.copy()

    for i,column_name in enumerate(['errors','warnings','internal_id']):
        if column_name not in df.columns:
            df.insert(loc=i,column=column_name,value='')

    #% mandatory columns validation
    mandatory_columns = [
        'errors',
        'warnings',
        'internal_id',
        'curve_date',
        'curve_ccy',
        'Δ_convention',
    ]
    df = df.dropna(axis=0, subset=mandatory_columns) # drop rows with blanks in mandatory columns
    missing_mandatory_columns = [col for col in mandatory_columns if col not in df.columns.to_list()]
    if len(missing_mandatory_columns) > 0:
        df['errors'] += f'missing mandatory columns: {missing_mandatory_columns}\n'
        return df

    # Validate  column data
    valid_field_values =  {'delta_convention': VALID_DELTA_CONVENTIONS,
                           'day_count_basis': VALID_DAY_COUNT_BASIS + [np.nan,'']}
    
    # Enforce to list of valid values
    for column_name in mandatory_columns:
        if column_name in valid_field_values.keys():
            bool_cond = df[column_name].isin(valid_field_values[column_name])
            df.loc[np.logical_not(bool_cond),'errors'] = 'invalid value for ' + column_name
    
    # fx option specific curve_ccy and delta_convention validation
    field = 'curve_ccy'
    bool_cond = df[field].isin(['usdaud','usdeur','usdgbp','usdnzd'])       
    df.loc[bool_cond,'warnings'] += field + ' value is not per common market convention\n' 

    bool_cond = np.logical_and(df['curve_ccy'].isin(['audusd','nzdeur','gbpusd','nzdusd']),df['Δ_convention'].str.contains('premium_adjusted'))
    df.loc[bool_cond,'warnings'] += 'the regular delta is the market delta_convention for this currency pair\n' 

    bool_cond = np.logical_and.reduce([
        np.logical_not(df['curve_ccy'].isin(['audusd','nzdeur','gbpusd','nzdusd','usdaud','usdeur','usdgbp','usdnzd'])),
        df['curve_ccy'].str.contains('usd'),
        df['Δ_convention'].str.contains('regular')])
    df.loc[bool_cond,'warnings'] += 'premium adjusted delta is the market delta_convention for this currency pair\n' 

    date_columns = ['tenor_name','tenor_date']
    optional_columns = ['day_count_basis','tenor_years','base_ccy','quote_ccy']

    for column_name in optional_columns:
        if column_name not in df.columns:
            df[column_name] = np.nan
        else:
            # Enforce to list of valid values
            if column_name in valid_field_values.keys():
                bool_cond = df[column_name].isin(valid_field_values[column_name])
                df.loc[np.logical_not(bool_cond),'errors'] = 'invalid value for ' + column_name

    user_input_columns = [v for v in df.columns if v not in mandatory_columns + date_columns + optional_columns]

    # drop user input columns if they are all nan
    mask = df[user_input_columns].isna().all()
    cols_to_drop = mask[mask].index.tolist()
    df = df.drop(columns=cols_to_drop)

    # Validate volatility input columns
    not_nan_user_input_columns = [v for v in df.columns if v not in mandatory_columns + date_columns + optional_columns]

    invalid_columns = []
    valid_volatility_input_columns = []
    for i,v in enumerate(not_nan_user_input_columns):
        pattern1 = r'^σ_(\d{1,2})Δ(call|put)$'
        pattern2 = r'^σ_(\d{1,2})Δ(bf|rr)$'
        atm_column_names = ['σ_atmΔneutral','σ_atmf']
        if (re.match(pattern1, v) and 1 <= int(re.match(pattern1, v).group(1)) <= 99) \
            or (re.match(pattern2, v) and 1 <= int(re.match(pattern2, v).group(1)) <= 99) \
            or v in atm_column_names:   
            valid_volatility_input_columns.append(v)
        else:
            invalid_columns.append(v)
            
    df = df.drop(columns=invalid_columns)
    
    for col in invalid_columns:
        msg = 'user added column' + "'" + col + "'" + ' does not ' \
            + 'match regex pattern ' + "'" + pattern1 + "', or pattern " + "'" + pattern2 + "'," \
            + ' and is not in the allowed list (' + ', '.join(atm_column_names) + ')\n'   
            
        bool_cond = df[col].isnotna()
        df.loc[bool_cond,'errors'] += msg

    # Enforce only one type of σ quote input
    pattern_call_put = r'^σ_(\d{1,2})Δ(call|put)$'
    pattern_strategy = r'^σ_(\d{1,2})Δ(bf|rr)$'
    cols_call_put = df.filter(regex=pattern_call_put).columns
    cols_strategy = df.filter(regex=pattern_strategy).columns
    mask_call_put = df[cols_call_put].apply(lambda x: x.notna().any(), axis=1)
    mask_strategy = df[cols_strategy].apply(lambda x: x.notna().any(), axis=1)
    if 'σ_atmf' in df.columns:
        mask_atmf = df['σ_atmf'].apply(lambda x: x.notna().any(), axis=1)
        
        mask = np.logical_and.reduce([mask_call_put, mask_strategy, np.logical_not(mask_atmf)])
        df.loc[mask,'errors'] = 'row has non-nan values for i) Δ-σ quotes and ii) σ-strategy quotes; specify only one volatility input type per row\n'

        mask = np.logical_and.reduce([mask_call_put, mask_atmf, np.logical_not(mask_strategy)])
        df.loc[mask,'errors'] = 'row has non-nan values for i) Δ-σ quotes and ii) σ-atmf quotes; specify only one volatility input type per row\n'                

        mask = np.logical_and.reduce([mask_strategy, mask_atmf, np.logical_not(mask_call_put)])
        df.loc[mask,'errors'] = 'row has non-nan values for i) σ-strategy quotes and ii) σ-atmf quotes; specify only one volatility input type per row\n'               
        
        mask = np.logical_and.reduce([mask_strategy, mask_atmf, mask_call_put])
        df.loc[mask,'errors'] = 'row has non-nan values for i) σ-strategy quotes and ii) σ-atmf quotes; specify only one volatility input type per row\n'      
        
        mask = mask_call_put & mask_strategy & mask_atmf
        df.loc[mask,'errors'] = 'row has non-nan values for across i) Δ-σ quotes, ii) σ-strategy quotes and iii) σ-atmf quotes; input only one volatility input type per row\n'
    else:
        mask = np.logical_and.reduce([mask_call_put, mask_strategy])
        df.loc[mask,'errors'] = 'row has non-nan values for i) Δ-σ quotes and ii) σ-strategy quotes; input only one volatility input type per row\n'
        
    # Enforce numeric types for strategy quotes
    for col in valid_volatility_input_columns:
        bool_cond = df[col].apply(lambda x: not isinstance(x, (float, int)) and not pd.isna(x))
        df.loc[bool_cond,'errors'] += col + ' has an invalid type\n' 
        
    # Call and put quotes validation
    for col in (cols_call_put.to_list() + atm_column_names):
        if col in df.columns:
            bool_cond = df[col].apply(lambda x: isinstance(x, (float, int)) and x <= 0.0)
            df.loc[bool_cond,'errors'] += col + ' must have a positive value\n'  
            
            bool_cond = df[col].apply(lambda x: isinstance(x, (float, int)) and x > 2) # unlikely for volatility quote to be > 200% 
            df.loc[bool_cond,'warnings'] += col + ' value is unusually large' 
     
    for col in cols_strategy.to_list():
        bool_cond = df[col].apply(lambda x: isinstance(x, (float, int)) and abs(x) > 0.25) # unlikely for strategy spread to be > 25%
        df.loc[bool_cond,'warnings'] += col + ' value is unusually large' 
            
    def extract_numbers(df, suffix):
        pattern = rf'^σ_(\d{{1,2}})Δ{suffix}$'
        cols = df.filter(regex=pattern).columns
        return [col.split('_')[1].split('Δ')[0] for col in cols]
    
    Δ_list = list(set(extract_numbers(df, 'rr') + extract_numbers(df, 'bf')))

    for Δ in Δ_list:
        atm = 'σ_atmΔneutral'
        bf = 'σ_' + Δ + 'Δbf'
        rr = 'σ_' + Δ + 'Δrr'
        
        for v in ['call','put']:
            column_name = 'σ_' + Δ + 'Δ' + v
            if column_name not in df.columns:
                df[column_name] = np.nan
        
        for i,row in df.iterrows():

            if bf and rr in row.index:     
                if atm in row.index:
                    if pd.notna(row[bf]) and pd.notna(row[rr]) and pd.notna(row[atm]):
                        if isinstance(row[bf], (float, int)) and isinstance(row[rr], (float, int)) \
                            and isinstance(row[atm], (float, int)) and row[atm] > 0:
                                                      
                                df.at[i,'σ_' + Δ + 'Δcall'] = row[bf] + row[atm] + 0.5 * row[rr]
                                df.at[i,'σ_' + Δ + 'Δput'] = row[bf] + row[atm] - 0.5 * row[rr]
               
                if pd.isna(row[bf]) and pd.notna(row[rr]):
                    df.loc[i,'errors'] += bf + ' value is absent\n' # add comment if butterfly is n/a
                elif pd.isna(row[rr]) and pd.notna(row[bf]):
                    df.loc[i,'errors'] += rr + ' value is absent\n' # add comment if risk reversal is n/a
                elif (pd.notna(row[bf]) or pd.notna(row[rr])) and pd.isna(row[atm]):
                    df.loc[i,'errors'] += atm + ' value is absent\n' # add comment if at-the-money is n/a
                    
            elif bf in row.index and rr not in row.index:
                if rr not in row.index and pd.notna(row[bf]):
                    df.loc[i,'errors'] += bf + ' value is present but column ' + rr + ' is absent\n'
                if atm not in row.index and pd.notna(row[bf]):
                    df.loc[i,'errors'] += bf + ' value is present but column ' + atm + ' is absent\n'                    
                    
            elif rr in row.index and bf not in row.index:
                if bf not in row.index and pd.notna(row[rr]):
                    df.loc[i,'errors'] += rr + ' value is present but column ' + bf + ' is absent\n'
                if atm not in row.index and pd.notna(row[rr]):
                    df.loc[i,'errors'] += rr + ' value is present but column ' + atm + ' is absent\n'                         
          
    # Drop σ-strategy quote columns
    pattern2 = r'^σ_(\d{1,2})Δ(bf|rr)$'
    cols_to_drop = df.filter(regex=pattern2).columns
    df = df.drop(columns=cols_to_drop)
    
    # Drop all nan columns
    df = df.dropna(axis=1, how='all')
        
    return df



    
    
    
    


# -*- coding: utf-8 -*-


if __name__ == "__main__":
    import os
    import pathlib
    os.chdir(pathlib.Path(__file__).parent.parent.parent.parent.resolve())     
    print('__main__ - current working directory:', os.getcwd())

from frm.frm.schedule.tenor import calc_tenor_date
from frm.frm.schedule.business_day_calendar import get_calendar
from frm.frm.schedule.daycounter import DayCounter

import numpy as np
import pandas as pd    


class Deposit():
        
    def __init__(self,  effective_date: pd.Timestamp, 
                        contractual_interest_rate_naac: float,
                        maturity_date: pd.Timestamp=None, 
                        tenor_name: str=np.nan,
                        day_count_basis: str='act/act',
                        local_currency_holidays=None,
                        city_holidays=None,
                        holiday_calendar=None
                        ):
    
        if holiday_calendar is None:
            holiday_calendar = get_calendar(local_currency_holidays, city_holidays)     
    
        if pd.isnull(maturity_date):
            maturity_date, tenor_name, _ = calc_tenor_date(effective_date, tenor_name, holiday_calendar=holiday_calendar, curve_ccy=local_currency_holidays)
        
        assert effective_date < maturity_date
        
        self.effective_date = effective_date
        self.maturity_date = maturity_date
        self.daycounter = DayCounter(day_count_basis)
        self.contractual_interest_rate_naac = contractual_interest_rate_naac
        self.tenor_name = tenor_name
        
    def implied_discount_factor(self):
        yrs = self.daycounter.year_fraction(self.effective_date, self.maturity_date)
        return 1 / (1 + self.interest_rate*yrs)
    
    def implied_cczr(self):
        yrs = self.daycounter.year_fraction(self.effective_date, self.maturity_date)
        return -np.log(1 / (1 + self.interest_rate*yrs)) / yrs
    








# -*- coding: utf-8 -*-
"""
Spyder Editor

This is a temporary script file.
"""

# https://gpt4all.io/index.html


import os
import pandas as pd
import numpy as np
import scipy as sc

os.chdir('C:/Users/-/OneDrive/Finance/OTC')

from otc.schedule.tenor import tenor_to_date_offset
from otc.calendars.calendar import get_calendar
from otc.schedule.daycounter import DayCounter



#%% Import and clean data

file_path = 'C:/Users//OneDrive/Finance/OTC/otc/notebooks/fx_data.xlsx'
sheet_name = 'fx_data_validation'  
df = pd.read_excel(file_path, sheet_name=sheet_name, usecols='A:H')

#%%

df['alpha'] = -1 * sc.stats.norm.ppf(0.25 * np.exp(df['base_ccy_depo_rates'] * df['years_to_maturity'])) 
df['K_atm'] = df['fx_spot'] * np.exp((df['base_ccy_depo_rates'] - df['term_ccy_depo_rates'] + 0.5 * df['σ_atm'] ** 2) * df['years_to_maturity']) 
df['K_25Δcall'] = df['fx_spot'] * np.exp(df['alpha'] * df['σ_25Δcall'] * np.sqrt(df['years_to_maturity']) + (df['base_ccy_depo_rates'] - df['term_ccy_depo_rates'] + 0.5 * df['σ_25Δcall'] ** 2) * df['years_to_maturity'])
df['K_25Δput'] = df['fx_spot'] * np.exp(-1 * df['alpha'] * df['σ_25Δput'] * np.sqrt(df['years_to_maturity']) + (df['base_ccy_depo_rates'] - df['term_ccy_depo_rates'] + 0.5 * df['σ_25Δput'] ** 2) * df['years_to_maturity'])
assert (df['K_25Δcall'] > df['K_atm']).all(), 'K_25Δcall is less than K_atm'
assert (df['K_25Δput'] < df['K_atm']).all(), 'K_25Δput is greater than than K_atm'
        
#%%


def d1(F,K,σ,t):
    dplus = (np.log(F/K) + 0.5 * (σ**2) * t) / (σ*np.sqrt(t))
    return dplus

def d2(F,K,σ,t):
    dminus = d1(F,K,σ,t) - σ * np.sqrt(t)
    return dminus


def VannaVolgaImpliedVol(F,K,t,K1,K2,K3,σ1,σ2,σ3):
    y1 = (np.log(K2/K) * np.log(K3/K)) / (np.log(K2/K1) * np.log(K3/K1))
    y2 = (np.log(K/K1) * np.log(K3/K)) / (np.log(K2/K1) * np.log(K3/K2))
    y3 = (np.log(K/K1) * np.log(K/K2)) / (np.log(K3/K1) * np.log(K3/K2))

    P = (y1*σ1 + y2*σ2 + y3 *σ3) - σ2
    
    Q = y1 * d1(F,K1,σ1,t) * d2(F,K1,σ1,t) * ((σ1-σ2)**2) \
      + y2 * d1(F,K2,σ2,t) * d2(F,K2,σ2,t) * ((σ2-σ2)**2) \
      + y3 * d1(F,K3,σ3,t) * d2(F,K3,σ3,t) * ((σ3-σ2)**2)
    
    d1d2 = d1(F,K,σ2,t) * d2(F,K,σ2,t)
    
    σ = σ2 + (-σ2 + np.sqrt(σ2**2 + d1d2 *(2*σ2*P+Q)))/(d1d2)
    
    return σ


#%%
strike = (1+np.arange(-0.20,0.21,0.001))*df['fx_spot'].iloc[0]

strikes_matrix = np.tile(strike,[len(df),1])

VVImpliedVol = np.zeros((len(df),len(strike)),dtype=float)
F = df['fx_spot'] * np.exp(np.multiply(df['base_ccy_depo_rates'] - df['term_ccy_depo_rates'],df['years_to_maturity']))

#%%

for i in range(len(strike)):
    VVImpliedVol[:,i]=VannaVolgaImpliedVol(F,strikes_matrix[:,i],df['years_to_maturity'],
        df['K_25Δput'], df['K_atm'], df['K_25Δcall'],
        df['σ_25Δput'], df['σ_atm'], df['σ_25Δcall'])
    
#%%
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm

X = df['years_to_maturity']
Y = strike
X, Y = np.meshgrid(X,Y)

Z = VVImpliedVol.T
fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
ax.set_xlabel('Time to maturity')
ax.set_ylabel('Strike')
ax.set_zlabel('Volatility')
surf = ax.plot_surface(X, Y, Z, cmap=cm.gnuplot,linewidth=0)
plt.show()




# -*- coding: utf-8 -*-

import numpy as np
from scipy.stats import norm



def simHeston(n, s0, v0, mu, kappa, theta, sigma, rho, delta, no=None, method=0):
    if no is None:
        no = np.random.normal(0, 1, (int(n/delta), 2))

    if len(no[:, 0]) != int(n/delta):
        raise ValueError('Size of no is inappropriate. Length of no[:, 0] should be equal to n/delta.')

    t = np.arange(0, np.ceil(n/delta)+1)
    sizet = len(t)
    x = np.zeros((sizet, 2))
    C = np.array([1, rho, rho, 1]).reshape(2, 2)
    u = normalcorr(C, no) * np.sqrt(delta)

    if method == 0:
        x[0, :] = [np.log(s0), v0]
        phiC = 1.5

        for i in range(1, sizet):
            m = theta + (x[i-1, 1] - theta) * np.exp(-kappa * delta)
            s2 = (x[i-1, 1] * sigma ** 2 * np.exp(-kappa * delta) / kappa * (1 - np.exp(-kappa * delta))
                  + theta * sigma ** 2 / (2 * kappa) * (1 - np.exp(-kappa * delta)) ** 2)
            phi = s2 / m ** 2
            gamma1, gamma2 = 0.5, 0.5
            K0 = -rho * kappa * theta / sigma * delta
            K1 = gamma1 * delta * (kappa * rho / sigma - 0.5) - rho / sigma
            K2 = gamma2 * delta * (kappa * rho / sigma - 0.5) + rho / sigma
            K3 = gamma1 * delta * (1 - rho ** 2)
            K4 = gamma2 * delta * (1 - rho ** 2)

            if phi <= phiC:
                b2 = 2 / phi - 1 + np.sqrt(2 / phi) * np.sqrt(2 / phi - 1)
                a = m / (1 + b2)
                x[i, 1] = a * (np.sqrt(b2) + no[i-1, 1]) ** 2
            else:
                p = (phi - 1) / (phi + 1)
                beta = (1 - p) / m

                if 0 <= norm.cdf(no[i-1, 1], 0, 1) <= p:
                    x[i, 1] = 0
                elif p < norm.cdf(no[i-1, 1], 0, 1) <= 1:
                    x[i, 1] = 1 / beta * np.log((1 - p) / (1 - norm.cdf(no[i-1, 1], 0, 1)))

            x[i, 0] = (x[i-1, 0] + mu * delta + K0 + K1 * x[i-1, 1] + K2 * x[i, 1]
                       + np.sqrt(K3 * x[i-1, 1] + K4 * x[i, 1]) * no[i-1, 0])

        x[:, 0] = np.exp(x[:, 0])
    else:
        x[0, :] = [s0, v0]

        for i in range(1, sizet):
            if method == 1:
                x[i-1, 1] = max(x[i-1, 1], 0)
            else:
                x[i-1, 1] = max(x[i-1, 1], -x[i-1, 1])

            x[i, 1] = (x[i-1, 1] + kappa * (theta - x[i-1, 1]) * delta
                       + sigma * np.sqrt(x[i-1, 1]) * u[i-1, 1])
            x[i, 0] = x[i-1, 0] + x[i-1, 0] * (mu * delta + np.sqrt(x[i-1, 1]) * u[i-1, 0])

    return x


def normalcorr(C, no):
    M = np.linalg.cholesky(C).T
    return (M @ no.T).T


def simGBM(n, x0, mu, sigma, delta, no=None, method=0):
    if no is None:
        no = np.random.normal(0, 1, int(np.ceil(n/delta)))
    else:
        no = no.flatten()
        if len(no) != int(np.ceil(n/delta)):
            raise ValueError('Error: length(no) <> n/delta')

    if method == 1: # Euler scheme
        x = x0 * np.cumprod(1 + mu * delta + sigma * delta**0.5 * no)
    elif method == 2: # Milstein scheme
        x = x0 * np.cumprod(1 + mu * delta + sigma * delta**0.5 * no +
                           0.5 * sigma**2 * delta * (no**2 - 1))
    elif method == 3: # 2nd order Milstein scheme
        x = x0 * np.cumprod(1 + mu * delta + sigma * delta**0.5 * no +
                           0.5 * sigma**2 * delta * (no**2 - 1) +
                           mu * sigma * no * (delta**1.5) + 0.5 * (mu**2) * (delta**2))
    else: # Direct integration
        x = x0 * np.exp(np.cumsum((mu - 0.5 * sigma**2) * delta + sigma * delta**0.5 * no))

    # Add starting value
    x = np.insert(x, 0, x0)
    return x



# -*- coding: utf-8 -*-


if __name__ == "__main__":
    import os
    import pathlib
    os.chdir(pathlib.Path(__file__).parent.parent.parent.parent.parent.resolve())     
    print('__main__ - current working directory:', os.getcwd())
    

import numpy as np
from numba import jit, prange
import scipy
from scipy.stats import norm
import timeit
import time
import pandas as pd
from frm.frm.market_data.ir_zero_curve import ZeroCurve
from frm.frm.pricing_engine.hull_white_1_factor import calc_theta
    
df = pd.read_excel('./frm/frm/pricing_engine/hull_white/zero_data.xlsx')

# running this - working through getting really clean functions
zc = ZeroCurve(curve_date=pd.Timestamp(2023,6,30),
               zero_data=df[['years','discount_factor']],
               interpolation_method='cubic_spline_on_zero_rates')

years = df['years']
years_linspace = np.linspace(min(years),max(years),100)

dt = 1e-5



f_plus_dt = zc.instantaneous_forward_rate(years=years_linspace+dt)
f_minus_dt = zc.instantaneous_forward_rate(years=years_linspace-dt)
df_dt = (f_plus_dt - f_minus_dt) / (2 * dt)
 
f = zc.instantaneous_forward_rate(years=years_linspace)

α = 0.19    # speed of convergence - Not required in Ho Lee Model
σ = 0.0196  # vol    

θ = df_dt + α * f + (σ**2)/(α*2)*(1-np.exp(-2*α*years_linspace))

theta = calc_theta(zc, α,σ,100)

# dt = 1e-5

# def forward_rate(tck, years):
    
#     scipy.interpolate.splev(x=years_linspace, tck=tck, der=1) # der=1 gets the 1st derivative



# f_t = 


# first_deriv = scipy.interpolate.splev(x=years_linspace, tck=tck, der=1) # der=1 gets the 1st derivative
# second_deriv = scipy.interpolate.splev(x=years_linspace, tck=tck, der=2) # der=1 gets the 1st derivative        



## Title:        All Functions
## Author:       Elisa FLeissner, Lars Stauffenegger
## Email:        elisa.fleissner@student.unisg.ch,
##               lars.stauffenegger@student.unisg.ch,
## Place, Time:  Zürich, 24.03.19
## Description:  All Functions for Option Pricing with Black Scholes, COS Method and Heston Model
## Improvements: -
## Last changes: -


# In[1]: Packages, Settings
import numpy as np
from scipy.special import erf


# In[2]: Standard Normal Cumulative Distribution Function
def StdNormCdf(z):
    phi = 0.5 * (1 + erf(z/np.sqrt(2)))
    return phi


# In[3]: Black Scholes Model
def blackScholes(S, K, r, tau, sigma, q):
    S = S * np.exp(-q * tau)
    d1 = np.divide((np.log(np.divide(S, K)) + (r + 1/2 * np.power(sigma, 2)) * tau), (sigma * np.sqrt(tau)))
    d2 = d1 - sigma * np.sqrt(tau)
    call  = np.multiply(S, StdNormCdf(d1)) - np.multiply(np.multiply(K, np.exp(-r * tau)), StdNormCdf(d2))
    put  = call + np.multiply(K, np.exp(-r * tau)) - S
    return call, put


# In[4]: Truncation Range based on Fitt et al. (2010)
def truncationRange(L, mu, tau, sigma, v_bar, lm, rho, volvol):
        c1 = mu * tau + (1 - np.exp(-lm * tau)) * (v_bar - sigma)/(2 * lm) - v_bar * tau / 2

        c2 = 1/(8 * np.power(lm,3)) * (volvol * tau * lm * np.exp(-lm * tau) \
            * (sigma - v_bar) * (8 * lm * rho - 4 * volvol) \
            + lm * rho * volvol * (1 - np.exp(-lm * tau)) * (16 * v_bar - 8 * sigma) \
            + 2 * v_bar * lm * tau * (-4 * lm * rho * volvol + np.power(volvol,2) + 4 * np.power(lm,2)) \
            + np.power(volvol,2) * ((v_bar - 2 * sigma) * np.exp(-2 * lm * tau) \
            + v_bar * (6 * np.exp(-lm * tau) - 7) + 2 * sigma) \
            + 8 * np.power(lm,2) * (sigma - v_bar) * (1 - np.exp(-lm * tau)))

        a = c1 - L * np.sqrt(np.abs(c2))
        b = c1 + L * np.sqrt(np.abs(c2))
        return a, b


# In[5]: Cosine Expansion based on Fang & Oosterlee (2008), translated from Prof. Gruber's Matlab Code
def cosSerExp(a, b, c, d, k):
    bma = b-a
    uu  = k * np.pi/bma
    chi = np.multiply(np.divide(1, (1 + np.power(uu,2))), (np.cos(uu * (d-a)) * np.exp(d) - np.cos(uu * (c-a)) * np.exp(c) + np.multiply(uu,np.sin(uu * (d-a))) * np.exp(d)-np.multiply(uu,np.sin(uu * (c-a))) * np.exp(c)))
    return chi


def cosSer1(a, b, c, d, k):
    bma    = b-a
    uu     = k * np.pi/bma
    uu[0]  = 1
    psi    = np.divide(1,uu) * ( np.sin(uu * (d-a)) - np.sin(uu * (c-a)) )
    psi[0] = d-c
    return psi


# In[6]: Characteristic Functions
def charFuncBSM(s, mu, sigma, T):
    phi = np.exp((mu - 0.5 * np.power(sigma,2)) * 1j * np.multiply(T,s) - 0.5 * np.power(sigma,2) * T * np.power(s,2))
    return phi

def charFuncHestonFO(mu, r, u, tau, sigma, v_bar, lm, rho, volvol):
    d = np.sqrt(np.power(lm - 1j * rho * volvol * u, 2) + np.power(volvol,2) * (np.power(u,2) + u * 1j))
    g = (lm - 1j * rho * volvol * u - d) / (lm - 1j * rho * volvol * u + d)
    C = np.divide(lm * v_bar, np.power(volvol,2)) * ( (lm - 1j * rho * volvol * u - d) * tau - 2 * np.log(np.divide((1 - g * np.exp(-d * tau)) , (1-g)) ))
    D = 1j * r * u * tau + sigma / np.power(volvol,2) * (np.divide((1 - np.exp(-d * tau)), (1 - g * np.exp(-d * tau)))) * (lm - 1j * rho * volvol * u - d) 
    phi = np.exp(D) * np.exp(C)
    return phi









# -*- coding: utf-8 -*-


if __name__ == "__main__":
    import os
    import pathlib
    os.chdir(pathlib.Path(__file__).parent.parent.parent.parent.parent.resolve())     
    print('__main__ - current working directory:', os.getcwd())
    
import numpy as np
from numba import jit, prange
import scipy
from scipy.stats import norm
import timeit
import time
import pandas as pd

from frm.frm.pricing_engine.monte_carlo_generic import generate_rand_nbs

import numpy.random as npr
from   scipy import integrate
import scipy.integrate as sciIntegr


import matplotlib.pyplot as plt



from frm.frm.pricing_engine.hull_white.Bond import *


class HullWhite(Bond):
    def __init__(self, ZCB, times, theta, kappa, sigma, r0=0.):
        Bond.__init__(self, theta, kappa, sigma, r0)
    
        self.dt     = 0.00001
        self._ZCB   = ZCB
        self._times = times
        
        self.ret = scipy.interpolate.splrep(self._times, -np.log(self._ZCB)/ self._times)                
                
    def ForwardRate(self, time):                
        r        = self.ZeroRate(time) # evaluate the zero rate 
        dr_dt    = scipy.interpolate.splev(time, self.ret, der=1) # der=1 gets the 1st derivative
         
        fwd_rate = r + time * dr_dt                    
        return fwd_rate
    
    def ZeroRate(self, time):            
        return scipy.interpolate.splev(time, self.ret)
    
    def PrepareThetas(self):
        Timeline        = np.linspace(Maturities[0], Maturities[-1],100)
        Thetarates      = [self.Thetat(t) for t in Timeline]
        self.ThetaratesForTimeline = np.array(Thetarates)
        self.Thetarates = scipy.interpolate.splrep(Timeline, Thetarates) 
        
    def A(self, t, T):                    
               
        def integrand1(time):
            theta = scipy.interpolate.splev(time, self.Thetarates)
            return theta * self.B(time, T)
            
        def integrand2(time):
            return (self.B(time, T)**2)
        
        val1 = integrate.quad(integrand1, t, T)[0]
        val2 = integrate.quad(integrand2, t, T)[0]
        return -val1 + 0.5*(self.sigma**2)*val2
        
    def Yield(self, t, T, rate):
        res    = -self.A(t, T)/(T-t) + (1/self.kappa)*(1-np.exp(-self.kappa*(T-t))) * 1/(T-t)*rate
        return res
        
    def Thetat(self, time):  
        
        # Calculate dF/dt by numerical integration
        up       = np.max(time + self.dt,0)
        down     = np.max(time - self.dt,0)
        
        df_t = self.ForwardRate(time)
        df_t_up = self.ForwardRate(up)  
        df_t_down = self.ForwardRate(down)

        df_dt    = (df_t_up - df_t_down)/(2*self.dt)
        
        return df_dt + self.kappa * df_t + (self.sigma**2)/(self.kappa*2)*(1-np.exp(-2*self.kappa*time))
        
    def ZCB_Forward_Integral(self,t, T):                                         
        val = integrate.quad(self.ForwardRate, t, T)[0]
        return np.exp(-val)    
    
    def B(self, t, T):
        return (1/self.kappa)*(1-np.exp(-self.kappa*(T- t)))
    
    def Exact_zcb(self, t, T):
        B = self.B(t, T)
        A = self.A(t, T)
        return np.exp(A-self.r0*B)
    
    def Euler(self, M, I, tau):
        # I is the number of simulation
        # M is the number of time steps until maturity
        # tau is the maturity        
        dt = tau / float(M)

        M  = int(M)
        I  = int(I)
        xh = np.zeros((M + 1, I))
        self.rates = np.zeros_like(xh)
        self.times = np.linspace(0, tau, num = M + 1)
        
        xh[0]     = self.r0
        for t in range(1, M + 1):
            xh[t] = xh[t - 1] + (self.Thetat(self.times[t-1]) - self.kappa*xh[t - 1]) * dt + \
                  self.sigma * np.sqrt(dt) * npr.standard_normal(I)
        self.rates = xh
        
    def ExpectedRate(self,t, T):
        # this expectation if always from time t = 0
        def integrand(time):
            theta = scipy.interpolate.splev(time, self.Thetarates)
            
            return theta * np.exp(-self.kappa*(T-time))
                    
        val = self.r0*np.exp(-(T-t)*self.kappa) + integrate.quad(integrand, t, T)[0]
        return val
    
    
#%%

dataframe = pd.read_csv('./frm/frm/pricing_engine/hull_white/Strips.csv', index_col=None)
print (dataframe.shape)
dataframe.head(10)


Maturities = np.asarray(dataframe.years)
Prices     = np.asarray(dataframe.discount_factor)
rate0      = -np.log(Prices[0])/Maturities[0]
print(rate0)

r0    = rate0   # current level of rates
kappa = 0.19    # speed of convergence - Not required in Ho Lee Model
theta = ""      # long term rate       - Not required in Ho Lee Model
sigma = 0.0196  # vol    

# create an instance of the object HullWhite and calculate theta(t)
hullwhite = HullWhite(Prices,Maturities, theta, kappa, sigma,rate0)
hullwhite.PrepareThetas()

theta = hullwhite.ThetaratesForTimeline

#%%

CCZR = -1 * np.log(Prices) / Maturities

spline_def = scipy.interpolate.splrep(Maturities, CCZR)   

Timeline        = np.linspace(Maturities[0], Maturities[-1],100)

CCZR_0 = scipy.interpolate.splev(Timeline, spline_def, der=0) # der=1 gets the 1st derivative
CCZR_1 = scipy.interpolate.splev(Timeline, spline_def, der=1) # der=1 gets the 1st derivative




#%%

print(-np.log(hullwhite.Exact_zcb(0,1)))
print(hullwhite.Yield(0,1,rate0))
print(-np.log(hullwhite.ZCB_Forward_Integral(0,1)))
print(dataframe.zero_rates[9])
print(dataframe.zero_rates[8])


#%%

# The forward rate for the period t = 1, t= 2 is 
print(-np.log(hullwhite.ZCB_Forward_Integral(1,2)))

# we can also use Hull and White ZCB formula to calculate the forward price for the period t = 1, t= 2 
print(hullwhite.Exact_zcb(0,2)/hullwhite.Exact_zcb(0,1))
print(hullwhite.ZCB_Forward_Integral(1,2))

Timeline  = np.linspace(Maturities[0], Maturities[-1],100)
Fwdrates  = [hullwhite.ForwardRate(t) for t in Timeline]
zerorates = [hullwhite.ZeroRate(t) for t in Timeline]
Thetarates= [hullwhite.Thetat(t) for t in Timeline]
hullwhitePrices=[hullwhite.Exact_zcb(0, t) for t in Timeline]

FwdratesNp = np.array(Fwdrates)
zeroratesNp = np.array(zerorates)
ThetaratesNp = np.array(Thetarates)
hullwhitePricesp = np.array(hullwhitePrices)


#%%

plt.figure(figsize=(16,4))
plt.subplot(121)
plt.plot(Timeline, Fwdrates, label ='calculated Forward rates')
plt.plot(Timeline, zerorates, label='calculated Spot rates')
plt.plot(Maturities, -np.log(Prices)/Maturities, marker='.', label='Original Spot rates')
plt.xlabel(r'Maturity $T$ ')
plt.title(r'continously compounded $r(t,T)$ and $f(t,T)$')
plt.grid(True)
plt.legend()
plt.ylabel('spot and forward rates')

plt.subplot(122)
plt.plot(Timeline, Thetarates, marker='.',label =r'$\theta(t)$')
plt.xlabel(r'Maturity $T$') 
plt.title(r'$function$ $\theta(t)$')
plt.grid(True)
plt.ylabel(r'$\theta(t)$')
plt.show()


#%%

plt.figure(figsize=(16,4))

plt.plot(Maturities, Prices, marker='*', label = "Original Prices")
plt.plot(Timeline, hullwhitePrices, label ='Hull White Prices')

plt.xlabel(r'Maturity $T$ ')
plt.title(r'Original Prices vs Hull White Model Recovered Prices')
plt.grid(True)
plt.legend()
plt.ylabel('Prices')

plt.show()


#%%

tau = 5
I   = 10000       # no. of simulations
M   = tau * 252   # trading day per annum

# So we always use the same random numbers
npr.seed(1500)

# run the sde
%time hullwhite.Euler(M, I, tau)

vals = hullwhite.StochasticPrice(hullwhite.rates, hullwhite.times)

print("the price is:      ", vals[0])
print("the price +2sd is :", vals[1])
print("the price -2sd is :", vals[2])
print("the Analytic price is :", hullwhite.Exact_zcb(0, 5))


#%%

# So we always use the same random numbers
npr.seed(1500)

size   = (Prices.shape[0])
Result = np.zeros((size, 6))
for i, j in enumerate(Maturities):
    tau = j
    M   = tau * 252
    hullwhite.Euler(M, I, tau)
    vals = hullwhite.StochasticPrice(hullwhite.rates, hullwhite.times)
    Result[i,0] = np.round(j,5)
    Result[i,1] = np.round(vals[0],4)
    Result[i,2] = np.round(vals[1],4)
    Result[i,3] = np.round(Prices[i],4)
    Result[i,4] = np.round(hullwhite.Exact_zcb(0,j),4)
    Result[i,5] = np.round(vals[2],4)
    
Result = pd.DataFrame(Result)
Result.columns = ['Maturity','MC Price','MC Price+2CD', 'Original Price', 'hullwhite Price', 'MC Price-2CD',]
Result.tail(10)


#%%

# So we always use the same random numbers
npr.seed(1500)

I   = 10000     # no. of simulations
T_O = 1 # 1 year
T_M = 5 # 5 year
M   = T_O * 252 # trading day per annum
    
BondValue = ZCBValue(0) # an instance of class ZCBValue, to model the payoff value at year 1    
vals = hullwhite.FutureZCB(M,I, T_O, T_M, BondValue)

#%%

print( "----------------------------------------------------------------")
print("the MC price of a ZCB maturing in",T_M, "years is:", np.round(vals[0],4))
print("The ZCB price from Formulae is               ", np.round(hullwhite.Exact_zcb(0, 5),4))

print( "----------------------------------------------------------------")
print("the MC price +2sd is :", vals[1])
print("the MC price -2sd is :", vals[2])
print( "----------------------------------------------------------------")

print("the MC price of a ZCB maturing in",T_O, "year is:", np.round(vals[4],4))
print("The ZCB price from Formulae is              ", np.round(hullwhite.Exact_zcb(0, 1),4))

print( "----------------------------------------------------------------")
print("The MC Expected short term rate in",T_O,"year is ", np.round(vals[5],4))
print("The Expected short term rate in",T_O,"year is ", np.round(hullwhite.ExpectedRate(0,1),4))

print( "----------------------------------------------------------------")
print("the Price in",T_O,"year of a ZCB price maturing in",T_M,"years is", np.round(vals[3],4))
print("The fwd price from Formulae is                           ", np.round(hullwhite.Exact_zcb(0, 5)/hullwhite.Exact_zcb(0, 1),4))


#%%

# So we always use the same random numbers
npr.seed(1500)

I   = 10000   # no. of simulations
OptionStrike = 0.8
BondValue    = ZCBOption(OptionStrike) # an instance of class ZCBOption, to model the payoff value at year 1   


#%%

%time vals = hullwhite.FutureZCB(M,I, T_O, T_M, BondValue)

#%%

print( "------------------------------------------------------")
print("the MC price of an one year Option on a ZCB maturing in:",T_M, "years is:", np.round(vals[0],4))

print( "------------------------------------------------------")
print("the MC price +1.96sd is :", vals[1])
print("the MC price -1.96sd is :", vals[2])

print("The Analytic hullwhite Price is ", 0.024040) # Hull White Analytic Option model is not included. Price was taken from
# Fixed Income Securities, Pietro Veronesi, chapter 19, example 19.4, page 662 


#%%


# -*- coding: utf-8 -*-


if __name__ == "__main__":
    import os
    import pathlib
    import sys
    os.chdir(pathlib.Path(__file__).parent.parent.parent.resolve())     
    print('__main__ - current working directory:', os.getcwd())

from typing import Literal

# https://www.iban.com/currency-codes
iban_ccy_codes = [
    ['AFGHANISTAN','Afghani','AFN',971],
    ['ALBANIA','Lek','ALL',8],
    ['ALGERIA','Algerian Dinar','DZD',12],
    ['AMERICAN SAMOA','US Dollar','USD',840],
    ['ANDORRA','Euro','EUR',978],
    ['ANGOLA','Kwanza','AOA',973],
    ['ANGUILLA','East Caribbean Dollar','XCD',951],
    ['ANTARCTICA','No universal currency','',],
    ['ANTIGUA AND BARBUDA','East Caribbean Dollar','XCD',951],
    ['ARGENTINA','Argentine Peso','ARS',32],
    ['ARMENIA','Armenian Dram','AMD',51],
    ['ARUBA','Aruban Florin','AWG',533],
    ['AUSTRALIA','Australian Dollar','AUD',36],
    ['AUSTRIA','Euro','EUR',978],
    ['AZERBAIJAN','Azerbaijanian Manat','AZN',944],
    ['BAHAMAS (THE)','Bahamian Dollar','BSD',44],
    ['BAHRAIN','Bahraini Dinar','BHD',48],
    ['BANGLADESH','Taka','BDT',50],
    ['BARBADOS','Barbados Dollar','BBD',52],
    ['BELARUS','Belarussian Ruble','BYN',933],
    ['BELGIUM','Euro','EUR',978],
    ['BELIZE','Belize Dollar','BZD',84],
    ['BENIN','CFA Franc BCEAO','XOF',952],
    ['BERMUDA','Bermudian Dollar','BMD',60],
    ['BHUTAN','Ngultrum','BTN',64],
    ['BHUTAN','Indian Rupee','INR',356],
    ['BOLIVIA (PLURINATIONAL STATE OF)','Boliviano','BOB',68],
    ['BOLIVIA (PLURINATIONAL STATE OF)','Mvdol','BOV',984],
    ['BONAIRE, SINT EUSTATIUS AND SABA','US Dollar','USD',840],
    ['BOSNIA AND HERZEGOVINA','Convertible Mark','BAM',977],
    ['BOTSWANA','Pula','BWP',72],
    ['BOUVET ISLAND','Norwegian Krone','NOK',578],
    ['BRAZIL','Brazilian Real','BRL',986],
    ['BRITISH INDIAN OCEAN TERRITORY (THE)','US Dollar','USD',840],
    ['BRUNEI DARUSSALAM','Brunei Dollar','BND',96],
    ['BULGARIA','Bulgarian Lev','BGN',975],
    ['BURKINA FASO','CFA Franc BCEAO','XOF',952],
    ['BURUNDI','Burundi Franc','BIF',108],
    ['CABO VERDE','Cabo Verde Escudo','CVE',132],
    ['CAMBODIA','Riel','KHR',116],
    ['CAMEROON','CFA Franc BEAC','XAF',950],
    ['CANADA','Canadian Dollar','CAD',124],
    ['CAYMAN ISLANDS (THE)','Cayman Islands Dollar','KYD',136],
    ['CENTRAL AFRICAN REPUBLIC (THE)','CFA Franc BEAC','XAF',950],
    ['CHAD','CFA Franc BEAC','XAF',950],
    ['CHILE','Unidad de Fomento','CLF',990],
    ['CHILE','Chilean Peso','CLP',152],
    ['CHINA','Yuan Renminbi','CNY',156],
    ['CHRISTMAS ISLAND','Australian Dollar','AUD',36],
    ['COCOS (KEELING) ISLANDS (THE)','Australian Dollar','AUD',36],
    ['COLOMBIA','Colombian Peso','COP',170],
    ['COLOMBIA','Unidad de Valor Real','COU',970],
    ['COMOROS (THE)','Comoro Franc','KMF',174],
    ['CONGO (THE DEMOCRATIC REPUBLIC OF THE)','Congolese Franc','CDF',976],
    ['CONGO (THE)','CFA Franc BEAC','XAF',950],
    ['COOK ISLANDS (THE)','New Zealand Dollar','NZD',554],
    ['COSTA RICA','Costa Rican Colon','CRC',188],
    ['CROATIA','Kuna','HRK',191],
    ['CUBA','Peso Convertible','CUC',931],
    ['CUBA','Cuban Peso','CUP',192],
    ['CURAÇAO','Netherlands Antillean Guilder','ANG',532],
    ['CYPRUS','Euro','EUR',978],
    ['CZECH REPUBLIC (THE)','Czech Koruna','CZK',203],
    ["CÔTE D'IVOIRE",'CFA Franc BCEAO','XOF',952],
    ['DENMARK','Danish Krone','DKK',208],
    ['DJIBOUTI','Djibouti Franc','DJF',262],
    ['DOMINICA','East Caribbean Dollar','XCD',951],
    ['DOMINICAN REPUBLIC (THE)','Dominican Peso','DOP',214],
    ['ECUADOR','US Dollar','USD',840],
    ['EGYPT','Egyptian Pound','EGP',818],
    ['EL SALVADOR','El Salvador Colon','SVC',222],
    ['EL SALVADOR','US Dollar','USD',840],
    ['EQUATORIAL GUINEA','CFA Franc BEAC','XAF',950],
    ['ERITREA','Nakfa','ERN',232],
    ['ESTONIA','Euro','EUR',978],
    ['ETHIOPIA','Ethiopian Birr','ETB',230],
    ['EUROPEAN UNION','Euro','EUR',978],
    ['FALKLAND ISLANDS (THE) [MALVINAS]','Falkland Islands Pound','FKP',238],
    ['FAROE ISLANDS (THE)','Danish Krone','DKK',208],
    ['FIJI','Fiji Dollar','FJD',242],
    ['FINLAND','Euro','EUR',978],
    ['FRANCE','Euro','EUR',978],
    ['FRENCH GUIANA','Euro','EUR',978],
    ['FRENCH POLYNESIA','CFP Franc','XPF',953],
    ['FRENCH SOUTHERN TERRITORIES (THE)','Euro','EUR',978],
    ['GABON','CFA Franc BEAC','XAF',950],
    ['GAMBIA (THE)','Dalasi','GMD',270],
    ['GEORGIA','Lari','GEL',981],
    ['GERMANY','Euro','EUR',978],
    ['GHANA','Ghana Cedi','GHS',936],
    ['GIBRALTAR','Gibraltar Pound','GIP',292],
    ['GREECE','Euro','EUR',978],
    ['GREENLAND','Danish Krone','DKK',208],
    ['GRENADA','East Caribbean Dollar','XCD',951],
    ['GUADELOUPE','Euro','EUR',978],
    ['GUAM','US Dollar','USD',840],
    ['GUATEMALA','Quetzal','GTQ',320],
    ['GUERNSEY','Pound Sterling','GBP',826],
    ['GUINEA','Guinea Franc','GNF',324],
    ['GUINEA-BISSAU','CFA Franc BCEAO','XOF',952],
    ['GUYANA','Guyana Dollar','GYD',328],
    ['HAITI','Gourde','HTG',332],
    ['HAITI','US Dollar','USD',840],
    ['HEARD ISLAND AND McDONALD ISLANDS','Australian Dollar','AUD',36],
    ['HOLY SEE (THE)','Euro','EUR',978],
    ['HONDURAS','Lempira','HNL',340],
    ['HONG KONG','Hong Kong Dollar','HKD',344],
    ['HUNGARY','Forint','HUF',348],
    ['ICELAND','Iceland Krona','ISK',352],
    ['INDIA','Indian Rupee','INR',356],
    ['INDONESIA','Rupiah','IDR',360],
    ['INTERNATIONAL MONETARY FUND (IMF) ','SDR (Special Drawing Right)','XDR',960],
    ['IRAN (ISLAMIC REPUBLIC OF)','Iranian Rial','IRR',364],
    ['IRAQ','Iraqi Dinar','IQD',368],
    ['IRELAND','Euro','EUR',978],
    ['ISLE OF MAN','Pound Sterling','GBP',826],
    ['ISRAEL','New Israeli Sheqel','ILS',376],
    ['ITALY','Euro','EUR',978],
    ['JAMAICA','Jamaican Dollar','JMD',388],
    ['JAPAN','Yen','JPY',392],
    ['JERSEY','Pound Sterling','GBP',826],
    ['JORDAN','Jordanian Dinar','JOD',400],
    ['KAZAKHSTAN','Tenge','KZT',398],
    ['KENYA','Kenyan Shilling','KES',404],
    ['KIRIBATI','Australian Dollar','AUD',36],
    ['KOREA (THE DEMOCRATIC PEOPLE’S REPUBLIC OF)','North Korean Won','KPW',408],
    ['KOREA (THE REPUBLIC OF)','Won','KRW',410],
    ['KUWAIT','Kuwaiti Dinar','KWD',414],
    ['KYRGYZSTAN','Som','KGS',417],
    ['LAO PEOPLE’S DEMOCRATIC REPUBLIC (THE)','Kip','LAK',418],
    ['LATVIA','Euro','EUR',978],
    ['LEBANON','Lebanese Pound','LBP',422],
    ['LESOTHO','Loti','LSL',426],
    ['LESOTHO','Rand','ZAR',710],
    ['LIBERIA','Liberian Dollar','LRD',430],
    ['LIBYA','Libyan Dinar','LYD',434],
    ['LIECHTENSTEIN','Swiss Franc','CHF',756],
    ['LITHUANIA','Euro','EUR',978],
    ['LUXEMBOURG','Euro','EUR',978],
    ['MACAO','Pataca','MOP',446],
    ['MADAGASCAR','Malagasy Ariary','MGA',969],
    ['MALAWI','Kwacha','MWK',454],
    ['MALAYSIA','Malaysian Ringgit','MYR',458],
    ['MALDIVES','Rufiyaa','MVR',462],
    ['MALI','CFA Franc BCEAO','XOF',952],
    ['MALTA','Euro','EUR',978],
    ['MARSHALL ISLANDS (THE)','US Dollar','USD',840],
    ['MARTINIQUE','Euro','EUR',978],
    ['MAURITANIA','Ouguiya','MRU',929],
    ['MAURITIUS','Mauritius Rupee','MUR',480],
    ['MAYOTTE','Euro','EUR',978],
    ['MEMBER COUNTRIES OF THE AFRICAN DEVELOPMENT BANK GROUP','ADB Unit of Account','XUA',965],
    ['MEXICO','Mexican Peso','MXN',484],
    ['MEXICO','Mexican Unidad de Inversion (UDI)','MXV',979],
    ['MICRONESIA (FEDERATED STATES OF)','US Dollar','USD',840],
    ['MOLDOVA (THE REPUBLIC OF)','Moldovan Leu','MDL',498],
    ['MONACO','Euro','EUR',978],
    ['MONGOLIA','Tugrik','MNT',496],
    ['MONTENEGRO','Euro','EUR',978],
    ['MONTSERRAT','East Caribbean Dollar','XCD',951],
    ['MOROCCO','Moroccan Dirham','MAD',504],
    ['MOZAMBIQUE','Mozambique Metical','MZN',943],
    ['MYANMAR','Kyat','MMK',104],
    ['NAMIBIA','Namibia Dollar','NAD',516],
    ['NAMIBIA','Rand','ZAR',710],
    ['NAURU','Australian Dollar','AUD',36],
    ['NEPAL','Nepalese Rupee','NPR',524],
    ['NETHERLANDS (THE)','Euro','EUR',978],
    ['NEW CALEDONIA','CFP Franc','XPF',953],
    ['NEW ZEALAND','New Zealand Dollar','NZD',554],
    ['NICARAGUA','Cordoba Oro','NIO',558],
    ['NIGER (THE)','CFA Franc BCEAO','XOF',952],
    ['NIGERIA','Naira','NGN',566],
    ['NIUE','New Zealand Dollar','NZD',554],
    ['NORFOLK ISLAND','Australian Dollar','AUD',36],
    ['NORTHERN MARIANA ISLANDS (THE)','US Dollar','USD',840],
    ['NORWAY','Norwegian Krone','NOK',578],
    ['OMAN','Rial Omani','OMR',512],
    ['PAKISTAN','Pakistan Rupee','PKR',586],
    ['PALAU','US Dollar','USD',840],
    ['PALESTINE, STATE OF','No universal currency','',],
    ['PANAMA','Balboa','PAB',590],
    ['PANAMA','US Dollar','USD',840],
    ['PAPUA NEW GUINEA','Kina','PGK',598],
    ['PARAGUAY','Guarani','PYG',600],
    ['PERU','Nuevo Sol','PEN',604],
    ['PHILIPPINES (THE)','Philippine Peso','PHP',608],
    ['PITCAIRN','New Zealand Dollar','NZD',554],
    ['POLAND','Zloty','PLN',985],
    ['PORTUGAL','Euro','EUR',978],
    ['PUERTO RICO','US Dollar','USD',840],
    ['QATAR','Qatari Rial','QAR',634],
    ['REPUBLIC OF NORTH MACEDONIA','Denar','MKD',807],
    ['ROMANIA','Romanian Leu','RON',946],
    ['RUSSIAN FEDERATION (THE)','Russian Ruble','RUB',643],
    ['RWANDA','Rwanda Franc','RWF',646],
    ['RÉUNION','Euro','EUR',978],
    ['SAINT BARTHÉLEMY','Euro','EUR',978],
    ['SAINT HELENA, ASCENSION AND TRISTAN DA CUNHA','Saint Helena Pound','SHP',654],
    ['SAINT KITTS AND NEVIS','East Caribbean Dollar','XCD',951],
    ['SAINT LUCIA','East Caribbean Dollar','XCD',951],
    ['SAINT MARTIN (FRENCH PART)','Euro','EUR',978],
    ['SAINT PIERRE AND MIQUELON','Euro','EUR',978],
    ['SAINT VINCENT AND THE GRENADINES','East Caribbean Dollar','XCD',951],
    ['SAMOA','Tala','WST',882],
    ['SAN MARINO','Euro','EUR',978],
    ['SAO TOME AND PRINCIPE','Dobra','STN',930],
    ['SAUDI ARABIA','Saudi Riyal','SAR',682],
    ['SENEGAL','CFA Franc BCEAO','XOF',952],
    ['SERBIA','Serbian Dinar','RSD',941],
    ['SEYCHELLES','Seychelles Rupee','SCR',690],
    ['SIERRA LEONE','Leone','SLL',694],
    ['SINGAPORE','Singapore Dollar','SGD',702],
    ['SINT MAARTEN (DUTCH PART)','Netherlands Antillean Guilder','ANG',532],
    ['SISTEMA UNITARIO DE COMPENSACION REGIONAL DE PAGOS "SUCRE"','Sucre','XSU',994],
    ['SLOVAKIA','Euro','EUR',978],
    ['SLOVENIA','Euro','EUR',978],
    ['SOLOMON ISLANDS','Solomon Islands Dollar','SBD',90],
    ['SOMALIA','Somali Shilling','SOS',706],
    ['SOUTH AFRICA','Rand','ZAR',710],
    ['SOUTH GEORGIA AND THE SOUTH SANDWICH ISLANDS','No universal currency','',],
    ['SOUTH SUDAN','South Sudanese Pound','SSP',728],
    ['SPAIN','Euro','EUR',978],
    ['SRI LANKA','Sri Lanka Rupee','LKR',144],
    ['SUDAN (THE)','Sudanese Pound','SDG',938],
    ['SURINAME','Surinam Dollar','SRD',968],
    ['SVALBARD AND JAN MAYEN','Norwegian Krone','NOK',578],
    ['SWAZILAND','Lilangeni','SZL',748],
    ['SWEDEN','Swedish Krona','SEK',752],
    ['SWITZERLAND','WIR Euro','CHE',947],
    ['SWITZERLAND','Swiss Franc','CHF',756],
    ['SWITZERLAND','WIR Franc','CHW',948],
    ['SYRIAN ARAB REPUBLIC','Syrian Pound','SYP',760],
    ['TAIWAN (PROVINCE OF CHINA)','New Taiwan Dollar','TWD',901],
    ['TAJIKISTAN','Somoni','TJS',972],
    ['TANZANIA, UNITED REPUBLIC OF','Tanzanian Shilling','TZS',834],
    ['THAILAND','Baht','THB',764],
    ['TIMOR-LESTE','US Dollar','USD',840],
    ['TOGO','CFA Franc BCEAO','XOF',952],
    ['TOKELAU','New Zealand Dollar','NZD',554],
    ['TONGA','Pa’anga','TOP',776],
    ['TRINIDAD AND TOBAGO','Trinidad and Tobago Dollar','TTD',780],
    ['TUNISIA','Tunisian Dinar','TND',788],
    ['TURKEY','Turkish Lira','TRY',949],
    ['TURKMENISTAN','Turkmenistan New Manat','TMT',934],
    ['TURKS AND CAICOS ISLANDS (THE)','US Dollar','USD',840],
    ['TUVALU','Australian Dollar','AUD',36],
    ['UGANDA','Uganda Shilling','UGX',800],
    ['UKRAINE','Hryvnia','UAH',980],
    ['UNITED ARAB EMIRATES (THE)','UAE Dirham','AED',784],
    ['UNITED KINGDOM OF GREAT BRITAIN AND NORTHERN IRELAND (THE)','Pound Sterling','GBP',826],
    ['UNITED STATES MINOR OUTLYING ISLANDS (THE)','US Dollar','USD',840],
    ['UNITED STATES OF AMERICA (THE)','US Dollar','USD',840],
    ['UNITED STATES OF AMERICA (THE)','US Dollar (Next day)','USN',997],
    ['URUGUAY','Uruguay Peso en Unidades Indexadas (URUIURUI)','UYI',940],
    ['URUGUAY','Peso Uruguayo','UYU',858],
    ['UZBEKISTAN','Uzbekistan Sum','UZS',860],
    ['VANUATU','Vatu','VUV',548],
    ['VENEZUELA (BOLIVARIAN REPUBLIC OF)','Bolivar','VEF',937],
    ['VENEZUELA (BOLIVARIAN REPUBLIC OF)','Bolivar','VED',926],
    ['VIET NAM','Dong','VND',704],
    ['VIRGIN ISLANDS (BRITISH)','US Dollar','USD',840],
    ['VIRGIN ISLANDS (U.S.)','US Dollar','USD',840],
    ['WALLIS AND FUTUNA','CFP Franc','XPF',953],
    ['WESTERN SAHARA','Moroccan Dirham','MAD',504],
    ['YEMEN','Yemeni Rial','YER',886],
    ['ZAMBIA','Zambian Kwacha','ZMW',967],
    ['ZIMBABWE','Zimbabwe Dollar','ZWL',932],
    ['ÅLAND ISLANDS','Euro','EUR',978]]

VALID_CCYS = {l[2] for l in iban_ccy_codes}



# -*- coding: utf-8 -*-


import os
import pathlib

if __name__ == "__main__":
    os.chdir(pathlib.Path(__file__).parent.parent.parent.resolve()) # path to ./frm/
    print(__file__.split('\\')[-1], os.getcwd())
    
from pricing_engine.garman_kohlhagen import garman_kohlhagen
import numpy as np
from scipy.stats import norm
from scipy.integrate import quad
from scipy.optimize import minimize, root
import time

def simHeston(n, s0, v0, mu, kappa, theta, sigma, rho, delta, no=None, method=0):
    
    
    # https://www.degruyter.com/document/doi/10.1515/math-2017-0058/html?lang=en
    # E + M scheme or QE scheme appear to be the quickest
    
    if no is None:
        no = np.random.normal(0, 1, (int(n/delta), 2))

    if len(no[:, 0]) != int(n/delta):
        raise ValueError('Size of no is inappropriate. Length of no[:, 0] should be equal to n/delta.')

    t = np.arange(0, np.ceil(n/delta)+1)
    sizet = len(t)
    x = np.zeros((sizet, 2))
    C = np.array([1, rho, rho, 1]).reshape(2, 2)
    u = normalcorr(C, no) * np.sqrt(delta)

    if method == 0:
        x[0, :] = [np.log(s0), v0]
        phiC = 1.5

        for i in range(1, sizet):
            m = theta + (x[i-1, 1] - theta) * np.exp(-kappa * delta)
            s2 = (x[i-1, 1] * sigma ** 2 * np.exp(-kappa * delta) / kappa * (1 - np.exp(-kappa * delta))
                  + theta * sigma ** 2 / (2 * kappa) * (1 - np.exp(-kappa * delta)) ** 2)
            phi = s2 / m ** 2
            gamma1, gamma2 = 0.5, 0.5
            K0 = -rho * kappa * theta / sigma * delta
            K1 = gamma1 * delta * (kappa * rho / sigma - 0.5) - rho / sigma
            K2 = gamma2 * delta * (kappa * rho / sigma - 0.5) + rho / sigma
            K3 = gamma1 * delta * (1 - rho ** 2)
            K4 = gamma2 * delta * (1 - rho ** 2)

            if phi <= phiC:
                b2 = 2 / phi - 1 + np.sqrt(2 / phi) * np.sqrt(2 / phi - 1)
                a = m / (1 + b2)
                x[i, 1] = a * (np.sqrt(b2) + no[i-1, 1]) ** 2
            else:
                p = (phi - 1) / (phi + 1)
                beta = (1 - p) / m

                if 0 <= norm.cdf(no[i-1, 1], 0, 1) <= p:
                    x[i, 1] = 0
                elif p < norm.cdf(no[i-1, 1], 0, 1) <= 1:
                    x[i, 1] = 1 / beta * np.log((1 - p) / (1 - norm.cdf(no[i-1, 1], 0, 1)))

            x[i, 0] = (x[i-1, 0] + mu * delta + K0 + K1 * x[i-1, 1] + K2 * x[i, 1]
                       + np.sqrt(K3 * x[i-1, 1] + K4 * x[i, 1]) * no[i-1, 0])

        x[:, 0] = np.exp(x[:, 0])
    else:
        x[0, :] = [s0, v0]

        for i in range(1, sizet):
            if method == 1:
                x[i-1, 1] = max(x[i-1, 1], 0)
            else:
                x[i-1, 1] = max(x[i-1, 1], -x[i-1, 1])

            x[i, 1] = (x[i-1, 1] + kappa * (theta - x[i-1, 1]) * delta
                       + sigma * np.sqrt(x[i-1, 1]) * u[i-1, 1])
            x[i, 0] = x[i-1, 0] + x[i-1, 0] * (mu * delta + np.sqrt(x[i-1, 1]) * u[i-1, 0])

    return x


def normalcorr(C, no):
    M = np.linalg.cholesky(C).T
    return (M @ no.T).T


def simGBM(n, x0, mu, sigma, delta, no=None, method=0):
    if no is None:
        no = np.random.normal(0, 1, int(np.ceil(n/delta)))
    else:
        no = no.flatten()
        if len(no) != int(np.ceil(n/delta)):
            raise ValueError('Error: length(no) <> n/delta')

    if method == 1: # Euler scheme
        x = x0 * np.cumprod(1 + mu * delta + sigma * delta**0.5 * no)
    elif method == 2: # Milstein scheme
        x = x0 * np.cumprod(1 + mu * delta + sigma * delta**0.5 * no +
                           0.5 * sigma**2 * delta * (no**2 - 1))
    elif method == 3: # 2nd order Milstein scheme
        x = x0 * np.cumprod(1 + mu * delta + sigma * delta**0.5 * no +
                           0.5 * sigma**2 * delta * (no**2 - 1) +
                           mu * sigma * no * (delta**1.5) + 0.5 * (mu**2) * (delta**2))
    else: # Direct integration
        x = x0 * np.exp(np.cumsum((mu - 0.5 * sigma**2) * delta + sigma * delta**0.5 * no))

    # Add starting value
    x = np.insert(x, 0, x0)
    return x


def pdfHestonInt(x, theta, kappa, sigma, rho, t, px):
    Gamma = kappa + complex(0,1) * rho * sigma * px
    Gamma2 = Gamma * Gamma
    Omega = np.sqrt(Gamma2 + sigma ** 2 * (px ** 2 - complex(0,1) * px))
    Omega2 = Omega * Omega
    Ft1 = kappa * theta / (sigma ** 2) * Gamma * t

    ncoeff = Omega2 - Gamma2 + 2 * kappa * Gamma
    coeff = ncoeff / (2 * kappa * Omega)
    Ot = Omega * t / 2
    Err = np.cosh(Ot)

    Ft2 = np.log(Err + coeff * np.sinh(Ot))
    Ft = Ft1 - 2 * kappa * theta / (sigma ** 2) * Ft2
    Eksp = np.exp(Ft + complex(0,1) * x * px)
    return np.real(Eksp)

def pdfHeston(x, theta, kappa, sigma, rho, t, par=0):
    n = len(x)
    y = np.zeros(n)
    pxm = 100

    for i in range(n):
        y[i] = 1 / (2 * np.pi) * quad(lambda s: pdfHestonInt(x[i], theta, kappa, sigma, rho, t, s), -100, 100, epsrel=1e-8)[0]

    if par == 1:
        import matplotlib.pyplot as plt
        plt.plot(x, y)
        plt.show()

    return y


def HestonVanilla(cp, s, k, v0, vv, rd, rf, tau, kappa, theta, lambda_, rho):
    def HestonVanillaInt(phi, m):
        a = kappa * theta
        u = [0.5, -0.5]
        b = [kappa + lambda_ - rho * vv, kappa + lambda_]
        x = np.log(s)
        y = np.log(k)

        d = np.sqrt((complex(0,1) * rho * vv * phi - b[m-1]) ** 2 - vv ** 2 * (2 * u[m-1] * phi * complex(0,1) - phi ** 2))
        g = (b[m-1] - rho * vv * phi * complex(0,1) - d) / (b[m-1] - rho * vv * phi * complex(0,1) + d)
        D = (b[m-1] - rho * vv * phi * complex(0,1) - d) / (vv ** 2) * ((1 - np.exp(-d * tau)) / (1 - g * np.exp(-d * tau)))
        C = (rd - rf) * phi * complex(0,1) * tau + a / (vv ** 2) * ((b[m-1] - rho * vv * phi * complex(0,1) - d) * tau - 2 * np.log((1 - g * np.exp(-d * tau)) / (1 - g)))
        f = np.exp(C + D * v0 + complex(0,1) * phi * x)
        F = np.real(np.exp(-complex(0,1) * phi * y) * f / (complex(0,1) * phi))

        return F

    P1 = 0.5 + 1/np.pi * quad(lambda phi: HestonVanillaInt(phi, 1), 0, np.inf, epsrel=1e-8)[0]
    P2 = 0.5 + 1/np.pi * quad(lambda phi: HestonVanillaInt(phi, 2), 0, np.inf, epsrel=1e-8)[0]

    Pplus = (1 - cp) / 2 + cp * P1
    Pminus = (1 - cp) / 2 + cp * P2

    P = cp * (s * np.exp(-rf * tau) * Pplus - k * np.exp(-rd * tau) * Pminus)
    return P



def HestonVanillaFitSmile(delta, marketvols, spot, rd, rf, tau, cp):

    strikes = garman_kohlhagen(S=spot,σ=marketvols,r_f=rf,r_d=rd,tau=tau,cp=cp,Δ=delta,task='strike',Δ_convention='regular_spot_Δ')
    nostrikes = len(strikes)
    v0 = (marketvols[nostrikes // 2])**2
    kappa = 1.5
    initparam = [2 * np.sqrt(v0), 2 * v0, 0]

    def HestonVanillaSSE(param, cp, spot, strikes, v0, marketvols, rd, rf, tau, kappa):
        number_calls = 1
        print(number_calls)
        
        vv, theta, rho = param
        if vv < 0 or theta < 0 or abs(rho) > 1:
            return np.inf
        P = np.zeros(nostrikes)
        IV = np.zeros(nostrikes)
        for i in range(nostrikes):
            t1 = time.time()
            P[i] = HestonVanilla(cp, spot, strikes[i], v0, vv, rd, rf, tau, kappa, theta, 0, rho)
            t2 = time.time()
            IV[i] = ImplVolFX(P[i], spot, strikes[i], rd, rf, tau, cp)
            t3 = time.time()
            print('HestonVanilla', t2-t1)
            print('ImplVol', t3-t2)
        return np.sum((marketvols - IV)**2)

    def ImplVolFX(X, S, K, rd, rf, tau, cp):
        
        return root(lambda vol: (garman_kohlhagen(S=spot,σ=vol,r_f=rf,r_d=rd,tau=tau,cp=cp,K=K,Δ=delta,task='px') - X), 0.001).x[0]
        #return root(lambda vol: (GarmanKohlhagen(S, K, vol, rd, rf, tau, cp, 0) - X), 0.001).x[0]

    res = minimize(lambda param: HestonVanillaSSE(param, cp, spot, strikes, v0, marketvols, rd, rf, tau, kappa), initparam)
    vv, theta, rho = res.x
    P = np.zeros(nostrikes)
    IV = np.zeros(nostrikes)
    
    t4 = time.time()
    print(t4-t3, 'res = minimize(lambda param')    
    
    for i in range(nostrikes):
        #t1 = time.time()
        P[i] = HestonVanilla(cp, spot, strikes[i], v0, vv, rd, rf, tau, kappa, theta, 0, rho)
        #t2 = time.time()
        #print(t2-t1, 'HestonVanilla(cp, spot, strikes[i], v0, vv, rd, rf, tau, kappa, theta, 0, rho)')                 
        IV[i] = ImplVolFX(P[i], spot, strikes[i], rd, rf, tau, cp)
        #t3 = time.time()
        #print(t3-t2,'IV[i] = ImplVolFX(P[i], spot, strikes[i], rd, rf, tau, cp)') 
        
    t5 = time.time()
    print(t5-t4,'for loop')  
    
    SSE = HestonVanillaSSE(res.x, cp, spot, strikes, v0, marketvols, rd, rf, tau, kappa)
    return v0, vv, kappa, theta, rho, IV, SSE





# -*- coding: utf-8 -*-
"""
Created on Fri Nov 29 11:33:46 2019

@author: SGFoste
"""


""" Bootstrapping the yield curve """
import math
    
class BootstrapYieldCurve(object):
    
    def __init__(self):
        self.zero_rates = dict()  # Map each T to a zero rate
        self.instruments = dict()  # Map each T to an instrument
        
    def add_instrument(self, par, T, coup, price,
                       compounding_freq=2):
        """  Save instrument info by maturity """
        self.instruments[T] = (par, coup, price, compounding_freq)
    
    def get_zero_rates(self):
        """  Calculate a list of available zero rates """
        self.__bootstrap_zero_coupons__()
        self.__get_bond_spot_rates__()
        return [self.zero_rates[T] for T in self.get_maturities()]
        
    def get_maturities(self):
        """ Return sorted maturities from added instruments. """
        return sorted(self.instruments.keys())
        
    def __bootstrap_zero_coupons__(self):
        """ Get zero rates from zero coupon bonds """
        for T in self.instruments.keys():
            (par, coup, price, freq) = self.instruments[T]
            if coup == 0:
                self.zero_rates[T] = \
                    self.zero_coupon_spot_rate(par, price, T)
                    
    def __get_bond_spot_rates__(self):
        """ Get spot rates for every marurity available """
        for T in self.get_maturities():
            instrument = self.instruments[T]
            (par, coup, price, freq) = instrument

            if coup != 0:
                self.zero_rates[T] = \
                    self.__calculate_bond_spot_rate__(
                        T, instrument)
                
    def __calculate_bond_spot_rate__(self, T, instrument):
        """ Get spot rate of a bond by bootstrapping """
        try:
            (par, coup, price, freq) = instrument
            periods = T * freq  # Number of coupon payments
            value = price
            per_coupon = coup / freq  # Coupon per period

            for i in range(int(periods)-1):
                t = (i+1)/float(freq)
                spot_rate = self.zero_rates[t]
                discounted_coupon = per_coupon * \
                                    math.exp(-spot_rate*t)
                value -= discounted_coupon

            # Derive spot rate for a particular maturity
            last_period = int(periods)/float(freq)        
            spot_rate = -math.log(value /
                                  (par+per_coupon))/last_period
            return spot_rate

        except:
            print("Error: spot rate not found for T=%s" % t)
            
    def zero_coupon_spot_rate(self, par, price, T):
        """ Get zero rate of a zero coupon bond """
        spot_rate = math.log(par/price)/T
        return spot_rate


if __name__ == "__main__":
    yield_curve = BootstrapYieldCurve()
    yield_curve.add_instrument(100, 0.25, 0., 97.5)
    yield_curve.add_instrument(100, 0.5, 0., 94.9)
    yield_curve.add_instrument(100, 1.0, 0., 90.)
    yield_curve.add_instrument(100, 1.5, 8, 96., 2)
    yield_curve.add_instrument(100, 2., 12, 101.6, 2)
    y = yield_curve.get_zero_rates()
    x = yield_curve.get_maturities()






import numpy as np
from scipy import integrate
import scipy.integrate as sciIntegr

class Payoff(object):
    def __init__(self, K):
        self.K = K
    def f(self):
        pass
    
class ZCBOption(Payoff):
    def f(self, S):
        return np.maximum(S - self.K,0)

class ZCBValue(Payoff):
    def f(self, S):
        return S

class Bond(object):
    def __init__(self, theta, kappa, sigma, r0=0.):
        '''
        r0    is the current level of rates
        kappa is the speed of convergence
        theta is the long term rate
        sigma is the volatility    
        '''
        self.theta = theta
        self.kappa = kappa
        self.sigma = sigma
        self.r0    = r0
        self.dt    = 0.001
    
    def B(self, t, T):
        pass
    
    def A(self, t, T): 
        pass
    
    def Exact_zcb(self, t, T):
        pass
    
    def Euler(self, M, I, T):
        pass
    
    def Yield(self, t, T, rate):
        return -1/(T-t)*self.A(t,T) + 1/(T-t)*self.B(t,T)*rate
    
    def SpotRate(self, t, T) :
        price = self.Exact_zcb(t, T)
        time  = T - t
        return (-np.log(price)/time)
    
    def ForwardRate1(self, time):
        up   = np.max(time + self.dt,0)
        down = np.max(time - self.dt,0)        

        r        = self.SpotRate(0, time)                        
        dr       = self.SpotRate(0, up) - self.SpotRate(0, down)
        dr_dt    = dr/(2*self.dt)         
        fwd_rate = r + time * dr_dt
        return fwd_rate

    def ZCB_Forward_Integral1(self, t, T):    
        val = integrate.quad(self.ForwardRate1, t, T)[0]
        return np.exp(-val)
    
    def ForwardRate2(self, time):            
        up   = np.max(time + self.dt,0)
        down = np.max(time - self.dt,0)

        dP       = self.Exact_zcb(0, down) - self.Exact_zcb(0, up)     
        dP_dt    = dP/(2*self.dt)
        P        = self.Exact_zcb(0, time)
        fwd_rate = dP_dt/P
        return fwd_rate

    def ZCB_Forward_Integral2(self, t, T):                    
        val = integrate.quad(self.ForwardRate2, t, T)[0]
        return np.exp(-val)

    def ForwardRate3(self, time):            
        up       = np.max(time + self.dt,0)
        down     = np.max(time - self.dt,0)

        dP       = np.log(self.Exact_zcb(0, down)) - np.log(self.Exact_zcb(0, up))     
        fwd_rate = dP/(2*self.dt)
        return fwd_rate 

    def ZCB_Forward_Integral3(self, t, T):                    
        val = integrate.quad(self.ForwardRate3, t, T)[0]
        return np.exp(-val)
    
    def StochasticPrice(self, VectorRates, VectorTime):  
        # VectorRates is a two dimensional array:
        # with simulated rates in columns and timesteps rates in rows
        
        # we do not need VectorRates and VectorTime at the beginning of the simulation as it is r0        
        VectorRates = VectorRates[:-1,:]
        VectorTime  = VectorTime[:-1]
        
        No_Sim = VectorRates.shape[1]
         
        price  = np.zeros(No_Sim)
        for i in range(No_Sim):    
            Rates    = VectorRates[:,i].T
            price[i] = np.exp(-(sciIntegr.simps(Rates , VectorTime)))        
        
        RangeUp_Down   = np.sqrt(np.var(price))*1.96 / np.sqrt(No_Sim)
        Mean = np.mean(price)
                
        return Mean,  Mean + RangeUp_Down, Mean - RangeUp_Down
        
    def FutureZCB(self,M, I, T_0, T_M, Bond):  
        
        self.Euler(M, I, T_0)        
        bond = Bond    
        
        # we do not need VectorRates and VectorTime at the beginning of the simulation as it is r0        
        VectorRates = self.rates[:-1,:]
        VectorTime  = self.times[:-1]
        
        No_Sim = VectorRates.shape[1]
         
        price          = np.zeros(No_Sim)
        priceUntil_FO  = np.zeros(No_Sim)
        Payoff         = np.zeros(No_Sim)
        R0             = np.zeros(No_Sim)
        for i in range(No_Sim):    
            Rates      = VectorRates[:,i].T
            R0[i]      = Rates[-1]
            Yield      = self.Yield(T_0, T_M, Rates[-1])
            ZCBPrice   = np.exp(-Yield*(T_M-T_0)) 
            Payoff[i]  = bond.f(ZCBPrice)    
            price[i]   = np.exp(-(sciIntegr.simps(Rates , VectorTime)))*Payoff[i]      
            priceUntil_FO[i]  = np.exp(-(sciIntegr.simps(Rates , VectorTime)))
        
        RangeUp_Down   = np.sqrt(np.var(price))*1.96 / np.sqrt(No_Sim)
        Mean     = np.mean(price)
        MeanRate = np.mean(R0)
        FWDValue = np.mean(Payoff)
        MeanValueUntil_FO = np.mean(priceUntil_FO)
        return Mean,  Mean + RangeUp_Down, Mean - RangeUp_Down, FWDValue, MeanValueUntil_FO, MeanRate
    
    def ExpectedRate(self,t, T):
        pass
    
    def VarianceRate(self,t, T):
        pass

    def PlotEulerSim(self,Text, No_of_Sim = 10):        
        # We plot the first No_of_Sim simulated paths 
        
        plt.plot(self.times, self.rates[:, :No_of_Sim], lw=1.5)
        plt.xlabel('time - yrs')  
        plt.ylabel('rates level')
        plt.grid(True)
        plt.title(Text + ' - the first {}'.format(No_of_Sim) + " Simulated Paths");
        plt.show()

    def PlotEulerSim_Stats(self, Text):        
        # We plot and compare the average simulation +-2 sd of all time steps vs. what we expect from the model
        
        SimAverage  = np.mean(self.rates, 1)
        SimSD       = np.sqrt(np.var(self.rates, 1))
        AnalyAverage= np.asarray([self.ExpectedRate(0, i) for i in self.times])
        AnalySD     = np.asarray([np.sqrt(self.VarianceRate(0, i)) for i in self.times])
        
        plt.plot(self.times, SimAverage,           lw=1.5, label ='Sim Mean',linestyle=':')
        #plt.plot(self.times, SimAverage + 2*SimSD, lw=1.5, label ='Sim Mean + 2*SD',linestyle=':')
        #plt.plot(self.times, SimAverage - 2*SimSD, lw=1.5, label ='Sim Mean - 2*SD',linestyle=':')
        
        plt.plot(self.times, AnalyAverage,             lw=1.5, label ='Analy Mean',linestyle='-')
        #plt.plot(self.times, AnalyAverage + 2*AnalySD, lw=1.5, label ='Analy Mean + 2*SD',linestyle='-.')
        #plt.plot(self.times, AnalyAverage - 2*AnalySD, lw=1.5, label ='Analy Mean - 2*SD',linestyle='-.')
                
        plt.legend()
        
        plt.xlabel('time - yrs')  
        plt.ylabel('rates')
        plt.grid(True)        
        plt.title(Text);        
        plt.show()


# -*- coding: utf-8 -*-

if __name__ == "__main__":
    import os
    import pathlib
    os.chdir(pathlib.Path(__file__).parent.parent.parent.resolve())     
    print('__main__ - current working directory:', os.getcwd())
        
from frm.frm.instruments.ir.swap import Swap
from frm.frm.instruments.ir.leg import Leg
from frm.frm.schedule.daycounter import DayCounter, VALID_DAY_COUNT_BASIS_TYPES
from frm.frm.market_data.iban_ccys import VALID_CCYS
from frm.frm.schedule.business_day_calendar import get_calendar
from frm.frm.schedule.tenor import calc_tenor_date
from frm.frm.utilities.utilities import convert_column_to_consistent_data_type

import scipy 
import pandas as pd
import numpy as np
from scipy.optimize import fsolve
from dataclasses import dataclass, field, InitVar
from typing import Optional, Union, Literal
import matplotlib.pyplot as plt
import datetime as dt
from dateutil.relativedelta import relativedelta

VALID_COMPOUNDING_FREQUENCY = Literal['continuously','simple','weekly','daily','monthly','quarterly','semiannually','annually']
VALID_INTERPOLATION_METHOD = Literal['linear_on_log_of_discount_factors','cubic_spline_on_zero_rates']
VALID_EXTRAPOLATION_METHOD = Literal['none','flat']

@dataclass
class ZeroCurve:
    # tk need to update this to work for (i) days, (ii) dates or (iii) years 
    curve_date: pd.Timestamp
    curve_ccy: str = None 
    
    historical_fixings: InitVar[Optional[pd.DataFrame]] = None
    day_count_basis: InitVar[Optional[VALID_DAY_COUNT_BASIS_TYPES]] = 'act/act'
    zero_data: InitVar[Optional[pd.DataFrame]] = None
    zero_rate_compounding_frequency: InitVar[VALID_COMPOUNDING_FREQUENCY] = 'continuously'
    interpolation_method: VALID_INTERPOLATION_METHOD = 'cubic_spline_on_zero_rates'
    extrapolation_method: VALID_EXTRAPOLATION_METHOD = 'none'
    instruments: InitVar[Optional[dict]] = None
    dsc_crv: InitVar[Optional['ZeroCurve']] = None
    fwd_crv: InitVar[Optional['ZeroCurve']] = None
    cubic_spline_definition = None
    
    # Non initialisation arguments
    daycounter: DayCounter = field(init=False)
    daily_data: pd.DataFrame = field(init=False, repr=False)
    
    def __post_init__(self, historical_fixings, day_count_basis, zero_data, zero_rate_compounding_frequency, instruments, dsc_crv, fwd_crv):
        
        if self.curve_ccy is not None:
            self.curve_ccy = self.curve_ccy.lower().strip()
            assert len(self.curve_ccy) in {3,6}
        
        self.daycounter = DayCounter(day_count_basis)
                
        if historical_fixings is not None:
            assert 'date' in historical_fixings.columns and 'fixing' in historical_fixings.columns 
            self.historical_fixings = dict(zip(historical_fixings.date, historical_fixings.fixing)) 
            
        if zero_data is not None:
            only_one_of_columns_X = ['days', 'tenor', 'years', 'date']
            only_one_of_columns_Y = ['zero_rate','discount_factor']
            
            if len(zero_data.columns.to_list()) != 2:
                raise ValueError('Exactly two columns must be specified: \n' / 
                                 + '(i) One of: ' + ', '.join(only_one_of_columns_X) + '\n' / 
                                 + '(ii) One of: ' + ', '.join(only_one_of_columns_Y))            
            
            specified_columns = [col for col in only_one_of_columns_X if col in zero_data.columns]
            if len(specified_columns) != 1:
                raise ValueError('Exactly one of the following columns must be specified: ' + ', '.join(only_one_of_columns_X))       
            
            
            specified_columns = [col for col in only_one_of_columns_Y if col in zero_data.columns]
            if len(specified_columns) != 1:
                raise ValueError('Exactly one of the following columns must be specified: ' + ', '.join(only_one_of_columns_Y))

            if 'tenor' in zero_data.columns:
                if len(self.curve_ccy) == 3:
                    holiday_calendar = get_calendar(ccys=[self.curve_ccy])
                elif len(self.curve_ccy) == 6:
                    holiday_calendar = get_calendar(ccys=[self.curve_ccy[:3],self.curve_ccy[3:]])
  
                date, tenor_cleaned, spot_date = calc_tenor_date(self.curve_date, zero_data['tenor'].values, self.curve_ccy, holiday_calendar)
                zero_data['tenor'] = tenor_cleaned
                zero_data['date'] = date
                zero_data['days'] = (zero_data['date'] - self.curve_date).days
                zero_data['years'] = self.daycounter.year_fraction(self.curve_date, zero_data['date'])

            elif 'date' in zero_data.columns:
                zero_data['days'] = (zero_data['date'] - self.curve_date).days
                zero_data['years'] = self.daycounter.year_fraction(self.curve_date, zero_data['date'])
                                
            elif 'days' in zero_data.columns:
                zero_data['date'] =  zero_data['days'].apply(lambda days: self.curve_date + dt.timedelta(days=days))
                zero_data['years'] = self.daycounter.year_fraction(self.curve_date, zero_data['date'])

            elif 'years' in zero_data.columns:
                pass
                                    
            # Clean up
            zero_data = zero_data.sort_values(by='years', ascending=True)       
            zero_data = zero_data.reset_index(drop=True)
            zero_data = convert_column_to_consistent_data_type(zero_data)
                
            if 'zero_rate' in zero_data.columns:
                zero_data['discount_factor'] = zero_rate_to_discount_factor(
                    years=zero_data['years'],
                    zero_rate=zero_data['zero_rate'],
                    zero_rate_compounding_frequency=self.zero_rate_compounding_frequency)

            elif 'discount_factor' in zero_data.columns:
                zero_data['zero_rate'] = discount_factor_to_zero_rate(
                    years=zero_data['years'],
                    discount_factor=zero_data['discount_factor'],
                    zero_rate_compounding_frequency=self.zero_rate_compounding_frequency)
                
            # Add nominal annual continuously compounded interest rate for internal use
            zero_data['nacc'] = -1 * np.log(zero_data['discount_factor']) / zero_data['years']
            zero_data.drop(columns=['zero_rate'], inplace=True)
            
            # Add pillar point at t=0 to improve interpolation if linear interpolation is applied
            # If cubic splines interpolation is applied this method can cause unstable splines
            # at the short end that affect the hull-white 1 factor theta, hence it is not applied. 
            # If, it is desired for completeness, the nacc, was set at t=0, based on the spline excluding this value.
            # This has not been tested but may work. 
            if self.interpolation_method == 'linear_on_log_of_discount_factors':
                first_row = {'years': 0.0, 'discount_factor': 1.0}
                first_row['nacc'] = zero_data['nacc'].iloc[0]    
                if 'tenor' in zero_data.columns:
                    first_row['tenor'] = ''
                if 'date' in zero_data.columns:
                    first_row['date'] = self.curve_date                
                if 'days' in zero_data.columns:
                    first_row['days'] = 0.0
                first_row_df = pd.DataFrame([first_row], columns=zero_data.columns)
                zero_data = pd.concat([first_row_df, zero_data])
                zero_data.reset_index(inplace=True, drop=True)            
            
            # Reorder the columns into a consistent format
            time_column_order = ['tenor','date','days','years','nacc','discount_factor']
            time_columns = [col for col in time_column_order if col in zero_data.columns]
            cols = time_columns + [col for col in zero_data.columns if col not in time_columns]
            zero_data = zero_data[cols]                         
            
            self.zero_data = zero_data
            self.__interpolate_daily_zero_data()
        else:
            # Or bootstrap zero curve from par instruments
            self.__bootstrap(instruments=instruments, dsc_crv=dsc_crv, fwd_crv=fwd_crv)
    

    def __interpolate_daily_zero_data(self):    
        
        if 'date' in self.zero_data.columns:
            max_date = max(self.zero_data['date'])
        else:
            max_years = int(max(self.zero_data['years']))
            max_date = self.curve_date + relativedelta(years=max_years)
            while self.daycounter.year_fraction(self.curve_date, max_date) < max_years:
                max_date += dt.timedelta(days=1)
                
        date_range = pd.date_range(self.curve_date,max_date,freq='d')
        days = self.daycounter.day_count(self.curve_date,date_range)
        years = self.daycounter.year_fraction(self.curve_date,date_range)
        
        if self.interpolation_method == 'cubic_spline_on_zero_rates':
            self.cubic_spline_definition = scipy.interpolate.splrep(x=self.zero_data['years'].to_list(),
                                                                    y=self.zero_data['nacc'].to_list(), k=3)  
            zero_rate_interpolated = scipy.interpolate.splev(years, self.cubic_spline_definition, der=0)
            self.zero_data_daily = pd.DataFrame({'date': date_range.to_list(), 
                                                 'days': days,
                                                 'years': years, 
                                                 'nacc': zero_rate_interpolated,
                                                 'discount_factor': np.exp(-1 * zero_rate_interpolated * years)})
            
        elif self.interpolation_method == 'linear_on_log_of_discount_factors':            
            ln_df_interpolated = np.interp(x=years, xp=self.zero_data['years'], fp= np.log(self.zero_data['discount_factor']))            
            self.zero_data_daily = pd.DataFrame({'date': date_range.to_list(),  
                                                 'days': days,
                                                 'years': years,
                                                 'nacc': -1 * ln_df_interpolated / years,
                                                 'discount_factor': np.exp(ln_df_interpolated)})        
       
    def __bootstrap(self, 
                  instruments: dict,
                  dsc_crv: Optional['ZeroCurve']=None,
                  fwd_crv: Optional['ZeroCurve']=None):
        
        # When dsc_crv is specified, we apply the dsc_crv to all relevent instruments and solve the fwd_crv that gives them a par value 
        # Example: 
        # Solving for the USD LIBOR 3M forward curve from USD LIBOR 3M fix-flt par swaps quoted from LCH (i.e that are fully collatarised). 
        # As these instruments are collatarised the discount curve does not equal the forward curve. 
        # We must solve this curve prior (i.e a USD SOFR curve or a cheapest-to-deliver OIS curve) and input it as the dsc_crv.
        
        # When fwd_crv is specified, we apply the fwd_crv to all relevent instruments and solve the dsc_crv that gives them a par value
        # Example:
        # Solving for the discount curve from USD floating rate bonds that reference USD LIBOR 1M
        # We must solve the USD LIBOR 1M curve prior and input this as the fwd_crv. 
        # As these instruments are collatarised the discount curve does not equal the forward curve. 
        # We must solve this discount curve prior (i.e a USD SOFR curve or a cheapest-to-deliver OIS curve)
        
        # This functionality will cover most most applications.  
        # This fuctionality does NOT cover the case you want to boostrap with 
        # a set of instruments reference this curve as the forward curve
        # with another set of instruments that reference this curve as the discount curve. 
        

        self.zero_data = {self.curve_date: 1.0}
        self.zero_data_daily = pd.DataFrame({'date': [self.curve_date],
                                     'years_to_date': [0.0], 
                                     'discount_factor': [1.0]})

        if 'deposits' in instruments.keys():
            deposits = instruments['deposits']
            deposits.sort(key=lambda x: x.maturity_date, reverse=False) 
            for deposit in instruments['deposits']:
                assert deposit.effective_date == self.curve_date, f"deposit.effective_date: {deposit.effective_date}, self.curve_date: {self.curve_date}"
                self.zero_data[deposit.maturity_date] = deposit.implied_discount_factor()
            self.__interpolate_zero_data()

        if 'futures' in instruments.keys():
            futures = instruments['futures']
            futures.sort(key=lambda x: x.maturity_date, reverse=False) 
            # Bootstrapping using futures/FRAs requires the futures to have overlap over the end and start dates of two successive futures
            # For example, a future ends on 15 June 2022, the next future must start on 15 June 2022 or earlier            
            for i,future in enumerate(futures):
                self.zero_data[future.maturity_date] = self.discount_factor(pd.Timestamp(future.effective_date)).at[0] \
                    / (future.forward_interest_rate * future.daycounter.year_fraction(future.effective_date,future.maturity_date) + 1)                                        
                self.__interpolate_zero_data()
        
        if 'FRAs' in instruments.keys():
            FRAs = instruments['FRAs']
            FRAs.sort(key=lambda x: x.maturity_date, reverse=False) 
            # Bootstrapping using futures/FRAs requires the futures to have overlap over the end and start dates of two successive futures
            # For example, a future ends on 15 June 2022, the next future must start on 15 June 2022 or earlier            
            for i,fra in enumerate(FRAs):
                self.zero_data[fra.maturity_date] = self.discount_factor([pd.Timestamp(future.effective_date)]).at[0] \
                    / (fra.forward_interest_rate * fra.daycounter.year_fraction(fra.effective_date,fra.maturity_date) + 1)                                        
                self.__interpolate_zero_data()
             
        if 'swaps' in instruments.keys():
            swaps = instruments['swaps']
            swaps.sort(key=lambda x: x.pay_leg.schedule['payment_date'].iloc[-1], reverse=False) 
            
            for i,swap in enumerate(swaps):
                final_payment_date = swap.pay_leg.schedule['payment_date'].iloc[-1]

                if final_payment_date > self.zero_data_daily['date'].iloc[-1]:
                    
                    # Set the initial guess to be the swaps par coupon rate
                    zero_rate = swap.pay_leg.fixed_rate if swap.pay_leg.leg_type == 'fixed' else swap.rec_leg.fixed_rate   
                    self.zero_data[final_payment_date] = np.exp(-zero_rate * self.daycounter.year_fraction(self.curve_date, final_payment_date))
                    
                    if dsc_crv is not None:
                        swap.set_discount_curve(dsc_crv)
                    if fwd_crv is not None:
                        swap.set_forward_curve(fwd_crv)
                    
                    def solve_to_par(zero_cpn_rate: [float],
                                     swap: Swap,
                                     zero_curve: ZeroCurve,
                                     final_payment_date: pd.Timestamp,
                                     solve_fwd_crv: bool,
                                     solve_dsc_crv: bool):
                        
                        zero_curve.zero_data[final_payment_date] = \
                            np.exp(-zero_cpn_rate[0] * zero_curve.daycounter.year_fraction(zero_curve.curve_date, final_payment_date))
                        zero_curve.__interpolate_zero_data()
              
                        if solve_fwd_crv:
                            swap.set_forward_curve(zero_curve)
                        if solve_dsc_crv:
                            swap.set_discount_curve(zero_curve)

                        pricing = swap.price()
                        return pricing['price']
                    
                    solve_fwd_crv = fwd_crv is None
                    solve_dsc_crv = dsc_crv is None    
                    x, infodict, ier, msg = fsolve(solve_to_par, [
                                                   zero_rate], 
                                                   args=(swap, self, final_payment_date, solve_fwd_crv, solve_dsc_crv), 
                                                   full_output=True)

                    self.zero_data[final_payment_date] = np.exp(-x[0] * self.daycounter.year_fraction(self.curve_date, final_payment_date))
                    self.__interpolate_zero_data()
                    
                    if ier != 1:
                        print('Error bootstrapping swap with maturity', swap.pay_leg.maturity_date.strftime('%Y-%m-%d'), msg)                        
                else:        
                    print('Swap with maturity', swap.pay_leg.maturity_date.strftime('%Y-%m-%d'), \
                          'excluded from bootstrapping as prior instruments have covered this period.')

        if 'legs' in instruments.keys():
            legs = instruments['legs']
            legs.sort(key=lambda x: x.schedule['payment_date'].iloc[-1], reverse=False) 
            
            for i,leg in enumerate(legs):
                final_payment_date = leg.schedule['payment_date'].iloc[-1]

                if final_payment_date > self.zero_data_daily['date'].iloc[-1]:
                    
                    # Set the initial guess to be the swaps par coupon rate
                    zero_rate = leg.fixed_rate if leg.leg_type == 'fixed' else leg.fixed_rate   
                    self.zero_data[final_payment_date] = np.exp(-zero_rate * self.daycounter.year_fraction(self.curve_date, final_payment_date))
                    
                    if dsc_crv is not None:
                        leg.set_discount_curve(dsc_crv)
                    if fwd_crv is not None:
                        leg.set_forward_curve(fwd_crv)
                    
                    def solve_to_par(zero_cpn_rate: [float],
                                     leg: Leg,
                                     zero_curve: ZeroCurve,
                                     final_payment_date: pd.Timestamp,
                                     solve_fwd_crv: bool,
                                     solve_dsc_crv: bool):
                        
                        zero_curve.zero_data[final_payment_date] = \
                            np.exp(-zero_cpn_rate[0] * zero_curve.daycounter.year_fraction(zero_curve.curve_date, final_payment_date))
                        zero_curve.__interpolate_zero_data()
              
                        if solve_fwd_crv:
                            leg.set_forward_curve(zero_curve)
                        if solve_dsc_crv:
                            leg.set_discount_curve(zero_curve)

                        
                        pricing = leg.price()
                        return pricing['price'] - leg.notional * (leg.transaction_price / 100.0)
                    
                    solve_fwd_crv = fwd_crv is None
                    solve_dsc_crv = dsc_crv is None    
                    x, infodict, ier, msg = fsolve(solve_to_par, [
                                                   zero_rate], 
                                                   args=(leg, self, final_payment_date, solve_fwd_crv, solve_dsc_crv), 
                                                   full_output=True)

                    self.zero_data[final_payment_date] = np.exp(-x[0] * self.daycounter.year_fraction(self.curve_date, final_payment_date))
                    self.__interpolate_zero_data()
                    
                    if ier != 1:
                        print('Error bootstrapping leg with maturity', leg.maturity_date.strftime('%Y-%m-%d'), msg)                        
                else:        
                    print('Leg with maturity', leg.maturity_date.strftime('%Y-%m-%d'), \
                          'excluded from bootstrapping as prior instruments have covered this period.')
    
    def flat_shift(self, 
                   basis_points: float=1) -> 'ZeroCurve':
        tenor_dates = list(self.zero_data.keys())
        years_to_date = self.daycounter.year_fraction(self.curve_date, tenor_dates)
        shifted = - np.log(list(self.zero_data.values())) / years_to_date + basis_points / 10000    
        df_shifted_data = pd.DataFrame({'date': tenor_dates, 
                                        'years_to_date': years_to_date, 
                                        'discount_factor':np.exp(-shifted * years_to_date)})
        df_shifted_data.loc[0,'discount_factor'] = 1.0 # set the discount factor on the curve date to be 1.0
        return ZeroCurve(curve_date=self.curve_date, 
                         zero_data = df_shifted_data,
                         day_count_basis = self.daycounter.day_count_basis)
                   
    def forward_rate(self,
                     d1: Union[pd.Timestamp, pd.Series],
                     d2: Union[pd.Timestamp, pd.Series],
                     compounding_frequency: VALID_COMPOUNDING_FREQUENCY='simple') -> pd.Series:

        assert len(d1) == len(d2)
        assert type(d1) == type(d2)
        
        bool_cond = d1 < self.curve_date
        
        historical_fixings = []
        if sum(bool_cond) > 0:
            historical_fixings = pd.Series([self.historical_fixings[d] if d in self.historical_fixings.keys() else np.nan for d in d1[bool_cond]])
            
        years = pd.Series(self.daycounter.year_fraction(d1, d2))            
        
        dsc_d1 = self.discount_factor(d1[np.logical_not(bool_cond)])
        dsc_d2 = self.discount_factor(d2[np.logical_not(bool_cond)])
        
        # https://en.wikipedia.org/wiki/Forward_rate
        if compounding_frequency == 'simple':
            forward_fixings = pd.Series(list((1 / years) * (dsc_d1 / dsc_d2 - 1)))
        elif compounding_frequency == 'annually':
            forward_fixings = pd.Series(list( (dsc_d1 / dsc_d2) ** (1/years)  - 1 ))
        elif compounding_frequency == 'continuously':
            forward_fixings = pd.Series(list((1 / years) * (np.log(dsc_d1) - np.log(dsc_d2))))
        
        forward_fixings[years==0] = 0
        
        return pd.Series(historical_fixings + forward_fixings.to_list())
       
    def instantaneous_forward_rate(self, years):        
        if self.interpolation_method == 'cubic_spline_on_zero_rates':
            zero_rate = self.zero_rate(years=years)
            zero_rate_1st_deriv = scipy.interpolate.splev(x=years, tck=self.cubic_spline_definition, der=1) 
            return zero_rate + years * zero_rate_1st_deriv
        else:
            raise ValueError('only supported for cubic spline interpolation method(s)')

    def discount_factor(self, 
                  dates: Optional[Union[pd.Timestamp, pd.Series]]=None,
                  days: Optional[Union[int, pd.Series]]=None,
                  years: Optional[Union[float, pd.Series]]=None,) -> pd.Series:
    
        inputs = {'dates': dates, 'days': days, 'years': years}
        if sum(x is not None for x in inputs.values()) != 1:
            raise ValueError('Only one input among days, date, or years is allowed.')            
        
        if years is not None:
            if self.interpolation_method == 'cubic_spline_on_zero_rates':
                nacc = scipy.interpolate.splev(years, self.cubic_spline_definition, der=0)
                discount_factor = np.exp(-1 * nacc * years)
                return discount_factor
    
            elif self.interpolation_method == 'linear_on_log_of_discount_factors':
                ln_df_interpolated = np.interp(x=years, xp=self.zero_data['years'], fp= np.log(self.zero_data['discount_factor']))
                return np.exp(ln_df_interpolated)
        else:
            if days is not None:
                dates = self.curve_date + dt.timedelta(days=days)
            if type(dates) is pd.Timestamp: 
                dates = pd.Series([dates])
            return pd.Series(pd.merge(pd.DataFrame(data=dates.values, columns=['date']), self.zero_data_daily, left_on='date',right_on='date')['discount_factor'].values)            
           
        
    def zero_rate(self, 
                  dates: Union[pd.Timestamp, pd.Series] = None,
                  days: Union[int, pd.Series] = None,
                  years: Union[float, pd.Series] = None,
                  compounding_frequency: VALID_COMPOUNDING_FREQUENCY='continuously') -> pd.Series:
                
        inputs = {'dates': dates, 'days': days, 'years': years}
        if sum(x is not None for x in inputs.values()) != 1:
            raise ValueError('Only one input among days, date, or years is allowed.')           
                
        if self.extrapolation_method == 'none':    
            msg = ". NaN may be returned. Consider applying the ZeroCurve 'extrapolation_method'"
        elif self.extrapolation_method == 'flat':
            msg = '. Flat extropolation will be applied.'

        if years is not None:
            discount_factor = self.discount_factor(dates=dates, days=days, years=years)
            zero_rate = discount_factor_to_zero_rate(years, 
                                                     discount_factor=discount_factor,
                                                     zero_rate_compounding_frequency=compounding_frequency)
            return zero_rate
        else:
            if days is not None:
                dates = self.curve_date + dt.timedelta(days=days)
            if type(dates) is pd.Timestamp: dates = pd.Series([dates])
            
            # check if any of the dates are outside the available data
            min_date = self.zero_data_daily['date'].min()
            below_range = dates < min_date
            dates_below = []
            if any(below_range):
                bound_date = self.zero_data_daily['date'].min()
                for i,date in enumerate(dates[below_range]):
                    print('Date', date.strftime('%Y-%m-%d'), 'is below the min available data of ' + bound_date.strftime('%Y-%m-%d') + msg)
                    if self.extrapolation_method == 'none':    
                        dates_below.append(np.nan)
                    elif self.extrapolation_method == 'flat':    
                        dates_below.append(bound_date)
            
            max_date = self.zero_data_daily['date'].max()
            above_range = dates > max_date
            dates_above = []
            if any(above_range):
                bound_date = self.zero_data_daily['date'].max()
                for i,date in enumerate(dates[above_range]):
                    print('Date', date.strftime('%Y-%m-%d'), 'is above the max available data of ' + bound_date.strftime('%Y-%m-%d') + msg)
                    if self.extrapolation_method == 'none':    
                        dates_above.append(np.nan)
                    elif self.extrapolation_method == 'flat':    
                        dates_above.append(bound_date)
          
            in_range = np.logical_and(np.logical_not(below_range),np.logical_not(above_range))
            cleaned_dates = dates_below + list(dates[in_range].values) + dates_above
            df = pd.merge(pd.DataFrame(data=cleaned_dates, columns=['date']), self.zero_data_daily, left_on='date',right_on='date', how='left')
            
            if compounding_frequency == 'continuously':
                return df['nacc'].values
            else: 
                zero_rate = discount_factor_to_zero_rate(years, 
                                                         discount_factor=df['discount_factor'], 
                                                         zero_rate_compounding_frequency=compounding_frequency)
                return zero_rate
                    
                    
    def plot(self, forward_rate_terms=[90]):
            
        # Zero rates
        min_date = self.zero_data_daily['date'].iloc[1]
        max_date = self.zero_data_daily['date'].iloc[-1]
        date_range = pd.date_range(min_date,max_date,freq='d')
        years_zr = self.daycounter.year_fraction(self.curve_date,date_range)
        zero_rates = pd.Series(self.zero_rate(dates=date_range)) * 100
        
        fig, ax = plt.subplots()
        
        ax.plot(years_zr, zero_rates, label='zero rate')  
  
        # Forward rates
        for term in forward_rate_terms:
            d1 = pd.date_range(min_date,max_date - pd.DateOffset(days=term),freq='d')
            d2 = pd.date_range(min_date + pd.DateOffset(days=term),max_date ,freq='d')
        
            years_fwd = self.daycounter.year_fraction(self.curve_date,d1)
            fwd_rates = pd.Series(self.forward_rate(d1,d2)) * 100       
        
            ax.plot(years_fwd, fwd_rates, label=str(term)+' day forward rate') 
        
        ax.set(xlabel='years', ylabel='interest rate (%)')
        
        ax.grid()
        ax.legend()
        plt.xlim([min(years_zr), 1.05*max(years_zr)])
        plt.show()

    def subtract_curve(self, 
                   zero_curve: 'ZeroCurve',
                   day_count_basis: Optional[VALID_DAY_COUNT_BASIS_TYPES]=None,
                   name: Optional[str]=None,
                   task=None):
    
        # Returns a zero curve constructed from zero rates where,
        # zero rates = self.zero_rates - zero_curve.zero_rates
        
        assert self.curve_date == zero_curve.curve_date        
        if day_count_basis is None:
            day_count_basis = self.daycounter.day_count_basis
            
        if name is None:
            if self.name is not None and zero_curve.name is not None:
                name = '[' + self.name + '] - [' + zero_curve.name + ']'
            
        min_date = min(self.zero_data_daily['date'].iloc[-1], zero_curve.daily_data['date'].iloc[-1])

        date_range = pd.date_range(self.curve_date, min_date,freq='d')
        date_range = date_range[1:] # exclude the zero_rate on self.data which is probably a np.nan

        zero_rates_self = self.zero_rate(dates=date_range)
        zero_rates_other = zero_curve.zero_rate(dates=date_range)
        
        zero_rate_difference = zero_rates_self - zero_rates_other
        df_diff = pd.DataFrame({'date': date_range,
                                'zero_rate': zero_rate_difference})
        
        return ZeroCurve(date=self.curve_date, 
                         name=name,
                         zero_data=df_diff,
                         day_count_basis=day_count_basis)
        
    
    def add_curve(self, 
                  zero_curve: 'ZeroCurve',
                  day_count_basis: Optional[VALID_DAY_COUNT_BASIS_TYPES]=None,
                  name: Optional[str]=None):
    
        # Returns a zero curve constructed from zero rates where,
        # zero rates = self.zero_rates + zero_curve.zero_rates
        
        assert self.curve_date == zero_curve.curve_date        
        if day_count_basis is None:
            day_count_basis = self.daycounter.day_count_basis
            
        if name is None:
            if self.name is not None and zero_curve.name is not None:
                name = '[' + self.name + '] + [' + zero_curve.name + ']'
            
        min_date = min(self.zero_data_daily['date'].iloc[-1], zero_curve.daily_data['date'].iloc[-1])

        date_range = pd.date_range(self.curve_date, min_date,freq='d')
        date_range = date_range[1:] # exclude the zero_rate on self.data which is probably a np.nan

        zero_rates_self = self.zero_rate(dates=date_range)
        zero_rates_other = zero_curve.zero_rate(dates=date_range)
        
        zero_rate_difference = zero_rates_self + zero_rates_other
        df_diff = pd.DataFrame({'date': date_range,
                                'zero_rate': zero_rate_difference})
        
        return ZeroCurve(date=self.curve_date, 
                         name=name,
                         zero_data=df_diff,
                         day_count_basis=day_count_basis)
        
        
    
def discount_factor_to_zero_rate(years, 
                                 discount_factor, 
                                 zero_rate_compounding_frequency: InitVar[VALID_COMPOUNDING_FREQUENCY] = 'continuously'):
    
    if zero_rate_compounding_frequency == 'continuously':
        return (- np.log(discount_factor) / years)
    elif zero_rate_compounding_frequency == 'daily':
        return (365 * ((1.0 / discount_factor) ** (1.0 / (365 * years)) - 1.0))
    elif zero_rate_compounding_frequency == 'monthly':
        return (12.0 * ((1.0 / discount_factor) ** (1.0 / (12.0 * years)) - 1.0))
    elif zero_rate_compounding_frequency == 'quarterly':
        return (4.0 * ((1.0 / discount_factor) ** (1.0 / (4.0 * years)) - 1.0))
    elif zero_rate_compounding_frequency == 'semi-annually':
        return (2.0 * ((1.0 / discount_factor) ** (1.0 / (2.0 * years)) - 1.0))
    elif zero_rate_compounding_frequency == 'annually':
        return ((1.0 / discount_factor) ** (1.0 / years) - 1.0)
    elif zero_rate_compounding_frequency == 'simple':
        return (((1.0 / discount_factor) - 1.0) / years)
    else:
        raise ValueError
    
def zero_rate_to_discount_factor(years, 
                                 zero_rate, 
                                 zero_rate_compounding_frequency: InitVar[VALID_COMPOUNDING_FREQUENCY] = 'continuously'):

    if zero_rate_compounding_frequency == 'continuously': 
        return np.exp(-zero_rate * years)
    elif zero_rate_compounding_frequency == 'daily': 
        return 1.0 / (1.0+zero_rate/365.0)**(365.0*years)
    elif zero_rate_compounding_frequency == 'monthly': 
        return 1.0 / (1.0+zero_rate/12.0)**(12.0*years)
    elif zero_rate_compounding_frequency == 'quarterly': 
        return 1.0 / (1.0+zero_rate/4.0)**(4.0*years)
    elif zero_rate_compounding_frequency == 'semi-annually': 
        return 1.0 / (1.0+zero_rate/2.0)**(2.0*years)
    elif zero_rate_compounding_frequency == 'annually': 
        return 1.0 / (1.0+zero_rate/1.0)**(1.0*years)
    else:
        raise ValueError

        
        
        
    